/*********************************************************************
 *              Copyright (c) 2021 PikeTec GmbH
 *               CONFIDENTIAL AND PROPRIETARY.
 *                   ALL RIGHTS RESERVED
 *              - PROPRIETARY OF PikeTec GmbH -
 **********************************************************************/
/* clang-format off */

/*
 * TPT VM Kernel interface
 * 
 * Use this API with caution:
 * ==========================
 * PikeTec tries to keep this interface as compatible as possible between TPT 
 * versions without causing users too much pain when migrating from one version to the next. 
 * Nevertheless, compatibility cannot always be 100% guaranteed. Therefore, all users of 
 * this API are advised to check compatibility and adjust if necessary before switching to 
 * a new version.
 */

#ifndef __TPT_VM_H_
#define __TPT_VM_H_

#ifdef __cplusplus
extern "C" {
#endif

/* ------------------------- */

#include "TPT_Types.h"
#include "TPT_Errors.h"


/*
 * This version number reflects the TPT "compatibilty version" of this API. Whenever this number changes there is a 
 * compatiblity change in the VM core API that does not guarantee backward compatibility any longer!
 * This number is incremented **only** with a new TPT version if the new TPT_VH.h API is no longer backward 
 * compatible with its previous version(s).
 * 
 * See `tpt_getVmCompatiblityVersion()`
 */
#define TPT_VM_COMPATIBLITY_VERSION 1

/*
 * Returns the "compatiblity version" this TPT VM binary has been compiled for. This number can differ from
 * TPT_VM_COMPATIBLITY_VERSION if the VM binary has been compiled with a different/backward-incompatible version.
 * 
 * To ensure 100% compatibility of user code with a VM binary, it is recommended to use `tpt_isCompatibleTo()`.
 */
TPT_DECLARE(tpt_uint16) tpt_getVmCompatiblityVersion();

/*
 * Returns the API "feature version" this API is compiled for. 
 * Note that the "feature version" number is only increased with a new TPT release if new features are added in this
 * release. Otherwise, the version number may remain unchanged.
 * 
 * In other words, if the function tpt_getVmFeatureVersion() returns the value 1800, it means that the current 
 * TPT version is **at least** TPT version 18.0.
 * 
 * Version History:
 * -----------------------------------------------------------------
 * - version 1800 (TPT 18)
 *       first TPT version supporting this constant 
 * -----------------------------------------------------------------
 * 
 * To ensure 100% compatibility of user code with a VM binary, it is recommended to use `tpt_isCompatibleTo()`.
 */
TPT_DECLARE(tpt_uint16) tpt_getVmFeatureVersion();

/*
 * This convenience macro can be used to check if this VM API is compatible to the specified TPT version.
 * 
 * if(tpt_isCompatibleTo(1800)) {
 *   // the VM binary is 100% compatible to the TPT18 API
 * }
 */
#define tpt_isCompatibleTo(featureversion) \
  ((tpt_getVmFeatureVersion() >= (featureversion) && (tpt_getVmCompatiblityVersion() == TPT_VM_COMPATIBLITY_VERSION))

/* each variable in the TPT VM has a particular role. */
typedef enum tpt_role_enum {
  tpt_input     = 0,   /* TPT input channel */
  tpt_output    = 1,   /* TPT output channel */
  tpt_param     = 2,   /* TPT parameter */
  tpt_local     = 3,   /* TPT local channel */
  tpt_vminternal= 4,   /* TPT VM internal variable (usually not visible/relevant for the user) */
  tpt_funcin    = 5,   /* function input argument */
  tpt_funcout   = 6,   /* function output argument and/or return value */
  tpt_assess    = 7,   /* TPT assessment variable (not in use within the TPT VM) */
  tpt_meas      = 8,   /* TPT measurement variable (not in use within the TPT VM) */
} tpt_role;


/*
 * an enum set is a collection of enum constants. Each constant has a name and a value,
 * both stored in the two arrays `names` and `value`.
 */
typedef struct tpt_enumset_struct {
  tpt_basetype  type;    /* primitive basetype of the enum set */
  tpt_uint16    size;    /* number of constants in this enum set */
  char**        names;   /* array with the enum constant names. The array contains `size` elements. */
  void*         value;   /* array with the enum constant values. Each values requires 1, 2, 4 or 8 bytes
                            (depending on unterlying type). Total length is therefore `size * bytesize` bytes. */
} tpt_enumset;


/* type forward declaration */
struct tpt_var_struct;


/*
 * function type of a function that is called if an input variable in a TPT VM cycle is determined lazily and
 * the value must be determined now (lazy read). Such a function can be specified per variable 
 * (see `tpt_var.value_cmd`). If called, the current value must be retrieved and stored/updated in 
 * `tpt_var.value`.
 *
 * @param var
 *      the variable tpt_var-struct to read. If the variable is an array the whole array (all elems) must be read
 *
 * @return
 *      `tpt_result_okay` if successful or `tpt_result_error` in case of an error.
 *      Error details can be retrieved using `tpt_getErrorCode()` and `tpt_getErrorMessage()`
 */
typedef tpt_result (*tpt_value_cmd) (struct tpt_var_struct *var);

/* 
 * special tag for the 'tptvar.dims' array. see documentation below 
 */
#define STRUCT_SELECTOR 0xFFFFFFFF

/* presentation of a single (flat) variable (channel, parameter) in the VM */
typedef struct tpt_var_struct {
  char*   name;                    /* variable name */
  tpt_uint16    id;                /* unique ID  */
  tpt_uint32    address;           /* start address of the variable value in internal VM memory */
  tpt_basetype  type;              /* base type  */
  tpt_uint8     bytesize;          /* byte width of a single element. Always 1,2, 4 or 8, depending on its 'type'. */
  tpt_tsize     dim;               /* total number of elements (> 0). for scalars, dim is 1 */
  void*         value;             /* current value (`dim * bytesize` bytes in total). Points to a data array used as 
                                      the "current value". Consider this pointer as a const pointer. So, do never
                                      change the pointer. Only the content */
  tpt_value_cmd value_cmd;         /* function for inputs that must be lazy loaded. NULL if the value is available
                                      initially. The lazy load function (if avail.) will be called per cycle as soon
                                      as the value is requested/read in the test model. The function should update the
                                      "value" area. */
  void*         value_cmd_arg;     /* additional optional custom argument for value_cmd-calls (optional, 
                                      by default it is NULL). This argument can be used by the platform to
                                      store additional context information for this variable (for constant
                                      time access at runtime). So basically it is a closure substitute for C. */
  tpt_bool      isQueued;          /* true if variable must be queued (Q) */
  tpt_bool      isNonQueued;       /* true if variable must be non-queued (NQ)  */
  tpt_bool      isRecord;          /* true if variable shall be recorded */
  tpt_bool      isConstant;        /* true if the variable is unchanged over time */
  tpt_bool      isString;          /* true if the variable is a string or an array of strings (base type is uint8 in this case) */
  tpt_bool      isHidden;          /* true if the variable has been marked as 'hidden' in the chosen mapping for the platform */
  tpt_bool      isReadWriteOutput; /* true if the `role` is `tpt_output`, `tpt_param`, or `tpt_funcout` but should be read from
                                      the SUT nonetheless (read/write semantics). */
  tpt_role      role;              /* role of the variable */
  tpt_double*   scalingCoeffs;     /* pointer to 4 scaling coeffs for `phys = (a * fix + b) / (c * fix + d)`
                                      or TPT_NULL if no scaling exists */
  char*         externalname;      /* external name or TPT_NULL if no external name exists, UTF-8 encoded */
  tpt_uint16    enumset;           /* index of the enum-Sets, this variable belongs to. 
                                      Value 0xffff if this is not an enum-variable */
  tpt_uint8     dims_size;         /* number of dimensions. `0` if this is a scalar (dim = 1 and dims == TPT_NULL in this case) */
  tpt_tsize*    dims;              /* If dims_size > 0, this array contains `dims_size` elements. It represents the individual 
                                      dimensions of multi-dimensional arrays, structs, and/or array-of-structs. 
                                      This array contains STRUCT_SELECTOR for all "positions" that represent a struct 
                                      element selector. 
                                      Example: 
                                      for "a[2].foo.bar[3]" the array contains the elements [2 0xFFFFFFFF 0xFFFFFFFF 3] */
  tpt_uint8     dims_reduced_size; /* number of dimensions 'n' in the `dims_reduced` array. 
                                       * `n=1` if this is a scalar , 
                                       * `n=1` if this is a non-array struct member (e.g. "foo.bar") , 
                                       * `n=1` if this is a 1-dim array. 
                                       * `n>1` if it is an n-dim array.
                                      This value is therefore always >= 1. */
  tpt_tsize*    dims_reduced;      /* This array contains `dims_reduced_size` elements. 
                                      It represents the individual dimensions of scalars, 1-dim array as well as
                                      multi-dim arrays. 
                                      - For scalars and non-array struct member (e.g. "foo.bar") it always contains [1].
                                      - For 1-dim arrays array[dim] it contains just the value "[dim]"
                                      - For n-dim arrays array[dim1][dim2]...[dimn] it contains the values 
                                        [dim1, dim2, ..., dimn]. */
  tpt_uint16    xaxis_id;          /* xaxis != 0xFFFF if this signal is a curve or map (refers to the xaxis variable) */
  tpt_uint16    yaxis_id;          /* yaxis != 0xFFFF if this signal is a map (refers to the yaxis variable) */
  char*         attributes;        /* string with additional attributes (key=value-pairs separated by \0). Ends with \0\0.
                                      TPT_NULL if there are no attributes */
} tpt_var;



/* pointer to a custom function struct */
typedef struct tpt_func_struct *tpt_func_ptr;



/*
 * Function handler type. A C function of this type is responsible for the processing of a corresponding TPT function.
 * 
 * Error-Handling:
 * If the function has a runtime error, the function must pass an error code and an error message to the VM before 
 * termination using `tpt_raiseError()`, so that the cause of the error is injected into TPT in the "normal" way.
 *
 * Normally `tpt_err_error_in_custom_function` should be passed as error code. However, another error code 
 * tpt_err_xxxx or an "arbitrary" error code" can be used, whose error code numeric value should be greater than 
 * `tpt_err_error_in_custom_function` (from this code on the error numbers are free).
 *
 * @param func
 *      the `tpt_func` struct of the function to be exeucted now.
 */
typedef void (*tpt_funcHandler) (tpt_func_ptr func);



/* representation of a single function (TPT server-stub function or TPT client function). */
typedef struct tpt_func_struct {
  tpt_uint16         id;                 /* unique function ID */
  char*              name;               /* function name */
  tpt_uint8          role;               /* 1 = server-stub function (implemented in TPT), 2 = client function */
  tpt_uint32         address;            /* entry point for server functions (program counter, internal purpose!) */
  tpt_uint16         parameters_size;    /* number of function args (inargs+outargs+returnarg) */
  tpt_var*           parameters;         /* array of function args (IN and OUT arguments as well as 
                                            return value, if any). The order is equal to the order of args. Optional
                                            return value is always the last in that list.
                                            Function arguments have the name "functionname:argname".
                                            The return value has the name "functionname::return" */
  tpt_funcHandler    handler;            /* points to the handler for this function which is called whenever this
                                            function is triggered at runtime. If the pointer is TPT_NULL (missing
                                            handler) a runtime error will be raised `tpt_err_customfunction_unknown` */
  void*              handlerCustomParam; /* additional optional custom argument for handler-calls (optional, 
                                            by default it is NULL). This argument can be used by the platform to
                                            store additional context information for this function (for constant
                                            time access at runtime). So basically it is a closure substitute for C. */
} tpt_func;



/* representation of an event */
typedef struct tpt_event_struct {
  tpt_uint16         id;                 /* unqiue event ID */
  char*              name;               /* logical name of an event */
} tpt_event;



/* representation of an event that has been triggered */
typedef struct tpt_event_call_struct {
  tpt_uint16         id;                 /* unique event ID, see `tpt_event.id` */
  tpt_uint16         data_size;          /* size of event data in bytes */
  tpt_byte*          data;               /* actual data blob of this event record */
  struct tpt_event_call_struct  *next;   /* linked list. refers to the next event in this list or TPT_NULL if this is the 
                                            last event */
} tpt_event_call;



/* 
 * This struct contains the data that the TPT VM exchanges with its environment. The struct can be 
 * accessed using `tpt_getData()`. Data must be copied from this struct (data is overwritten during 
 * the next `tpt_runCycle()`).
 */
typedef struct tpt_data_struct {
  tpt_uint16       enumsets_size;        /* number of enum sets */
  tpt_enumset*     enumsets;             /* array of enum sets */
  tpt_uint16       vars_size;            /* number of all variables (channels, params, function args) */
  tpt_var*         vars;                 /* array of all variables (channels, params, function args) */
  tpt_uint16       inputs_size;          /* Anzahl der Eingangskanäle (Teilmenge der "vars") */
  tpt_var*         inputs;               /* Pointer auf das Array der Eingangskanäle (Teilmenge der "vars") */
  tpt_uint16       outputs_size;         /* Anzahl der Ausgangskanäle (Teilmenge der "vars") */
  tpt_var*         outputs;              /* Pointer auf das Array der Ausgangskanäle (Teilmenge der "vars") */
  tpt_uint16       locals_size;          /* Anzahl der lokalen Kanäle (Teilmenge der "vars") */
  tpt_var*         locals;               /* Pointer auf das Array der lokalen Kanäle (Teilmenge der "vars") */
  tpt_uint16       funcs_size;           /* number of functions (server and client functions) */
  tpt_func*        funcs;                /* array of functions (server and client functions) */
  tpt_uint16       events_size;          /* number of event types */
  tpt_event*       events;               /* array of event types */
  tpt_event_call*  tpt_event_calls;      /* linked list of actual event records. TPT_NULL if no event recorded yet */
} tpt_data;



/* function type that is used for implementing the message box step for the TPT VM. see `tpt_installMessageBoxCmd()` */
typedef tpt_uint8 (*tpt_messageBoxCommand)(tpt_uint32  id,         /* unique id of the message box */
                                           tpt_bool    cancel,     /* true if msgbox should be cancelled/disposed now */
                                           char*       msg,        /* actual msg as HTML text */
                                           tpt_uint8   noButtons,  /* number of buttons to show */
                                           char*       buttons);   /* list of button texts, separated by \0, terminated by \0\0 */

/*
 * function to determine the current (real) time in seconds [s] since the test has been started.
 * This function is important if TPT tests run in realtime mode (platform dependent).
 * See `tpt_installRealtimeCmd()`
 */
typedef tpt_double (*tpt_realtimeCommand)();



/*
 * Loading a program. The program must be contained in the byte array 'program'. The function does not 
 * change the contents of program, but it also does not copy all data from this program. Therefore, the
 * program must remain at the given address after calling `tpt_load()` for all subsequent calls (e.g. of 
 * 'tpt_runCycle()`). 
 * 
 * This function allows to load and execute different programs one after the other (but always sequentially).
 * To restart a program, call the function `tpt_load()` several times.
 * 
 * The function returns 'tpt_result_okay' if the program was loaded properly. Otherwise the function 
 * returns `tpt_result_error`. The error message and code can be read with `tpt_getErrorMessage()` and 
 * `tpt_getErrorCode()`.
 */
TPT_DECLARE(tpt_result) tpt_load(tpt_byte program[]);



/*
 * returns the sample rate for test execution in micro seconds (10^-6 seconds).
 *
 * After `tpt_load()` the sample rate is the sample rate stored in the program. 
 * The value can still be changed using `tpt_setDensity()` as long as at most one cycle has been executed
 * using `tpt_runCycle()` in this program.
 */
TPT_DECLARE(tpt_uint32) tpt_getDensity();

/*
 * same as tpt_getDensity() but returns the density in seconds as double. Note that this double 
 * value may have numerical inaccuracies.
 */
#define tpt_getDensityInSec() ((tpt_double)vmstate->density / (tpt_double)1000000)



/*
 * The predefined sample rate (as specified in the program) can be changed using this function as 
 * long as at most one cycle has been executed in this program.
 *
 * The function returns `tpt_result_okay` if the value could be applied and `tpt_result_error` if
 * the program has been started already (and the error code of `tpt_getErrorCode()` is 
 * `tpt_err_density_frozen`).
 */
TPT_DECLARE(tpt_result) tpt_setDensity(tpt_uint32 samplerate);



/*
 * returns the number of cycles to be buffered for signals mit arbitrary time access (e.g. "signal(t-2*x)").
 * This value is stored in the program.
 */
TPT_DECLARE(tpt_uint32) tpt_getHistorySize();



/*
 * returns the maximum number of cycles to be executed until the program stops reporting a timeout.
 * This value is stored in the program which has been loaded using `tpt_load()` before.
 * 
 * Example:
 * If timeout == 3, then the VM will return `tpt_result_terminated` after the 3rd call of `tpt_runCycle()`.
 * Nevertheless, this third cycle is still executed normally to check whether the program still 
 * terminates on its own.
 */
TPT_DECLARE(tpt_uint32) tpt_getTimeout();



/*
 * The predefined timeout can be changed using this function. It can also changed while the program is
 * running already. The function always returns `tpt_result_okay`.
 */
TPT_DECLARE(tpt_result) tpt_setTimeout(tpt_uint32 timeout);



/*
 * returns the struct containing the actual data that the TPT VM exchanges with its environment.
 */
TPT_DECLARE(tpt_data*) tpt_getData();

/**
  * returns <code>tpt_true</code> if the data of a cycle should be recorded and otherwise 
  * <code>tpt_false</code>. This flag is controlled by the measurement() feature in TPT, which can be used to 
  * decide whether or not to save data for an individual cycle or not.
  */
TPT_DECLARE(tpt_bool) tpt_isRecordCycle();

/*
 * executes a single step of the previously loaded program. Before calling this function, all 
 * input channels and parameters must be initialized with their "current values".
 * After the call, values of the output channels can be read.
 *
 * (1) The function returns `tpt_result_okay` if the calculation terminates properly. 
 * (2) The function returns `tpt_result_terminated` if the program has reached the final
 *     state of the test case. In this case the output channels are no longer defined 
 *     after calling the function.
 * (3) The function returns `tpt_result_error' if a runtime error occurred during the 
 *     calculation of the program. In this case the error message can be retrieved 
 *     using `tpt_getErrorCode()` and `tpt_getErrorMessage()`.
 */
TPT_DECLARE(tpt_result) tpt_runCycle();



/*
 * calls the TPT server stub function with ID "id" (see `tpt_func.id`). Usually the user of this
 * API calls this method from within its "SUT code", i.e. between subsequent calls of `tpt_runCycle()'
 * to compute some intermeadiate results of the TPT VM.
 * 
 * (1) The function returns `tpt_result_okay` if the calculation terminates properly. 
 * (2) The function returns `tpt_result_error` if a runtime error occurred during the 
 *     calculation of the TPT function "id". In this case the error message can be 
 *     retrieved using `tpt_getErrorCode()` and `tpt_getErrorMessage()`.
 */
TPT_DECLARE(tpt_result) tpt_runServerStubFunction(tpt_uint16 id);



/*
 * returns the number of performed `tpt_runCylce()` calls after `tpt_load()`. 
 * Immediately after `tpt_load()` this function returns 0. After each call of `tpt_runCycle()` the
 * counter increments by 1.
 */
TPT_DECLARE(tpt_uint32) tpt_getCycleCount();



/*
 * true if the variable at the given address has been written/updated during the most recent VM cycle. Note
 * that the actual signal value may be unchanged. The mere fact that it has been written is sufficient.
 *
 * This flag will be set also for lazy inputs (inputs with "value_cmd") if the value has been read during
 * the most recent VM cycle.
 */
TPT_DECLARE(tpt_bool) tpt_isWrittenFlag(tpt_uint32 address);



/*
 * Returns the total length of the program. (This information is inherently stored within program
 * and retrieved by this function). If the array does *not* contain a TPT program or the
 * passed program is not in the correct version -1 is returned.
  */
TPT_DECLARE(tpt_int32) tpt_getProgramLength(tpt_byte prog[]);


/*
 * Specifies whether the data is to be recorded SUT-synchronously.
 * This information is read from the program file that has been loaded before (using 'tpt_load()').
 */
TPT_DECLARE(tpt_bool) tpt_isSutSynchron();



/*
 * installs the function responsible for processing message boxes. By default no such function exists 
 * (equivalent to TPT_NULL), so no message boxes can be used. If no message box command is installed, no
 * message box may be used in the program. Otherwise a runtime error will occur during 'tpt_runCycle()' with
 * error code `tpt_err_no_msgbox_handler`.
 */
TPT_DECLARE(void) tpt_installMessageBoxCmd(tpt_messageBoxCommand msgBoxCmd);



/*
 * installs the realtime command if the real time is defined "externally". By default the realtime command is
 * TPT_NULL. In this case it is assumed that the real time is identical to the simulation time (i.e. always a
 * multiple of the configured stepsize).
 */
TPT_DECLARE(void) tpt_installRealtimeCmd(tpt_realtimeCommand rtCmd);

/**
 * delivers a pointer to a (read only) string which contains all "suppressed signal names" (comma separated). The
 * string is empty if no signals have been suppressed.
 * 
 * Suppressed signals are those signals which are declared as inputs or outputs in the interface, but which are
 * not used in the current test case and are therefore suppressed due to the "Use effective interface" option.
 */
TPT_DECLARE(const char*) tpt_getSuppressedSignalNames();


/**
 * returns true if the signal with the given name has been suppressed.
 * 
 * Suppressed signals are those signals which are declared as inputs or outputs in the interface, but which are
 * not used in the current test case and are therefore suppressed due to the "Use effective interface" option.
 */
TPT_DECLARE(tpt_bool) tpt_isSuppressedSignalName(const char *signalname);

#ifdef __cplusplus
}
#endif


#endif /* __TPT_VM_H_ */
