/*********************************************************************
 *              Copyright (c) 2022 PikeTec GmbH
 *               CONFIDENTIAL AND PROPRIETARY.
 *                   ALL RIGHTS RESERVED
 *              - PROPRIETARY OF PikeTec GmbH -
 **********************************************************************/
/* clang-format off */

/**
 * @file
 * @brief This interface defines the connection to the TPT-VM..
 */

#ifndef __TPT_VM_API_H_
#define __TPT_VM_API_H_

#include "TPT_Types.h"
#include "TPT_Errors.h"

#ifdef __cplusplus
extern "C"
{
#endif

  /* ========================== TYPES & CONSTANTS ============================== */

  /** 
   * Handle to the tpt-vm as delivered by tpt_vmapi_createExecution() 
   */
  typedef void *tpt_vmapi_handle;

  /** 
   * Function handler type for client functions (i.e. functions implemented in the SUT code
   * that are callable from the TPT byte code)
   */
  typedef void (*tpt_vmapi_funchandler) (tpt_uint16 id, const char* name);

  /**
   * Function which delivers current time stamp in seconds.
   * This function is called in case the functionality is enabled in TPT.
   */
  typedef tpt_double (*tpt_vmapi_realtimeCommand)();

  /**
   * handle to a TPT VM signal 
   */
  typedef void* tpt_vmapi_signalhandle;
  
  /**
   * function pointer used for functions that allow "dynamic" access to SUT variables. "Dynamic" means that
   * the address does not have to be provided as a constant "value" address when calling tptvmapi_bindSignal_v2() but
   * may be initialized later or even changed at test runtime. Usually this mechanism is used if the
   * address of a variable is still unknown initially when tpt_vmapi_bindSignal_v2() binds C variables to TPT signals.
   * In this case, the API function tpt_vmapi_bindSignalDynamic_v2() has to be used instead of
   * tpt_vmapi_bindSignal_v2().
   *
   * The TPT-VM will retrieve the C-variable address of a signal using a valueaccess_fcn in every cycle
   * of the test execution (both, for reading and writing values). If such a valueaccess_fcn returns
   * a nullptr as the variable's address, the reading/writing of the corresponding variable is omitted for
   * the current cycle. (Note that addresses are often unknown initially at t=0s and, therefore, no
   * value can be copied from/to the C variable. In this case, the valueaccess_fcn simply returns nullptr and
   * the TPT VM will omit the signal for the current cycle.)
   *
   * @param idx
   *      index to be accessed (if the variable is an array). The index is 0 if the variable is NOT an array.
   *      If the variable is multi-dimensional, idx represents the "flattened index": Assume that the overall
   *      dimension consists of var[dim_1][dim_2]...[dim_n-1][dim_n] and the actual access happens at indices
   *      var[i_1][i_2]...[i_n-1][i_n], then the flattened index is
   *      idx = (i_1)*(dim_2)*...*(dim_n)
   *          + (i_2)*(dim_3)*...*(dim_n)
   *          + (i_3)*(dim_4)*...*(dim_n)
   *          + ...
   *          + (i_n-1)*(dim_n)
   *          + (i_n)
   * @param dims_size
   *      number of dimensions. For scalars this value is 0, for (flat) arrays the value is 1, for matrices the value
   *      is 2 and so on.
   * @param dims
   *      constant array of dimensions that represent the dimension of the signal in TPT. The number of value stored
   *      in this array is always dims_size. For multi-dimensional arrays this information might be relevant to determine
   *      the individual array coefficients i_1, ..., i_n by inverting the computation described above
   */
  typedef void* (*tpt_vmapi_valueaccess_fcntype_v2)(tpt_tsize idx, tpt_uint8 dims_size, const tpt_tsize *dims);


   /**
   * Function pointer used for functions that allow access to SUT data via getter and setter. This means that
   * instead of providing a fixed pointer to a variable a getter and setter function can be provided. This way any 
   * C-Code value can be accessed, even if the compiler does not allow accessing their address (e.g. for bit-fields).
   * In this case, the API function tpt_vmapi_bindSignalGetSet_v2() has to be used instead of
   * tpt_vmapi_bindSignal_v2().
   *
   * The TPT-VM will call the configured getter/setter functions in every cycle of the test execution (both, for 
   * reading and writing values). Both functions (getter and setter) shall return tpt_true in case the operation 
   * was successfully executed and tpt_false otherwise.
   *
   * @param valPtr Pointer to the address where the value shall be written to (in case of a getter function) 
   *       or read from (in case of a setter function).
   *       The underlying data type will allways correspond to the fixBaseType passed to the 
   *       tpt_vmapi_bindSignalGetSet_v2 function. So this pointer can safely be casted to a pointer of that data type.
   * @param idx
   *      index to be accessed (if the variable is an array). The index is 0 if the variable is NOT an array.
   *      If the variable is multi-dimensional, idx represents the "flattened index": Assume that the overall
   *      dimension consists of var[dim_1][dim_2]...[dim_n-1][dim_n] and the actual access happens at indices
   *      var[i_1][i_2]...[i_n-1][i_n], then the flattened index is
   *      idx = (i_1)*(dim_2)*...*(dim_n)
   *          + (i_2)*(dim_3)*...*(dim_n)
   *          + (i_3)*(dim_4)*...*(dim_n)
   *          + ...
   *          + (i_n-1)*(dim_n)
   *          + (i_n)
   * @param dims_size
   *      number of dimensions. For scalars this value is 0, for (flat) arrays the value is 1, for matrices the value
   *      is 2 and so on.
   * @param dims
   *      constant array of dimensions that represent the dimension of the signal in TPT. The number of value stored
   *      in this array is always dims_size. For multi-dimensional arrays this information might be relevant to determine
   *      the individual array coefficients i_1, ..., i_n by inverting the computation described above
   */
  typedef tpt_bool (*tpt_vmapi_valuegetset_fcntype_v2)(void* valPtr, tpt_tsize idx, tpt_uint8 dims_size, const tpt_tsize *dims);


  /**
   *  phys = (slope * fix + bias)/(slopeInv * fix + biasInv).
   * "no scaling" means coeffs {1,0,0,1}
   */
  typedef struct tpt_vmapi_scaling {
    tpt_double slope;    /* slope value (factor) used for scaling (scaling from SUT-fixed to TPT-phys) */
    tpt_double bias;     /* bias value (offset) used for scaling (scaling from SUT-fixed to TPT-phys) */
    tpt_double slopeInv; /* slopeInv value (factor) used for scaling (scaling from SUT-fixed to TPT-phys) */
    tpt_double biasInv;  /* biasInv value (offset) used for scaling (scaling from SUT-fixed to TPT-phys) */
  } tpt_vmapi_scaling;

  /** 
   * min/max limits; TPT_NAN represents "no limit".
   */
  typedef struct tpt_vmapi_limits {
    tpt_double minLimit; /* may be TPT_NAN */
    tpt_double maxLimit; /* may be TPT_NAN */
  } tpt_vmapi_limits;

  /**
   * This struct summarizes the basic propoerties of signals that have been bound using "tpt_vmapi_bindSignal_v2()",
   * "tpt_vmapi_bindSignalDynamic_v2()", or "tpt_vmapi_bindSignalGetSet_v2()". Each struct serves as a linked list:
   * the "next" element points to the next signal until nullptr has been reached.
   */
  typedef struct tpt_vmapi_signalconfig {
    const char*       name;              /* signal name */
    tpt_basetype      physType;          /* physical type (type of the signal in TPT) */
    tpt_basetype      fixType;           /* fix point type (type of the signal in C code) */
    tpt_vmapi_scaling scaling;           /* scaling coeffs. Coeffs {1,0,0,1} represent "no scaling". */
    tpt_vmapi_limits  limits;            /* physical limits. Use value TPT_NaN for "no limit". */
    tpt_tsize         dim;               /* total dimension of this signal. 1 for scalars. For multidim arrays this is 
                                            the product of all dims. */
    tpt_uint32*       dims;              /* If dims_size > 0, this array contains `dims_size` elements. It represents the individual
                                            dimensions of multi-dimensional arrays, structs, and/or array-of-structs.
                                            This array contains 0xFFFFFFFF for all "positions" that represent a struct
                                            element selector ('.').
                                            Example:
                                            For "a[2].foo.bar[3]" the array contains the elements [2 0xFFFFFFFF 0xFFFFFFFF 3]
                                            Note that this information is not available during parameter exchange (i.e. when
                                            running the TPT VM API with "-r" parameter) */
    tpt_uint8         dims_size;         /* number of dimensions. `0` if this is a scalar (dim = 1 and dims == TPT_NULL in this case)
                                            Note that this information is not available during parameter exchange (i.e. when
                                            running the TPT VM API with "-r" parameter) */
    tpt_bool          isInput;           /* true if signal is an input.  
                                            It can be both input and output if tpt_vmapi_enableReadWrite() has been called. */
    tpt_bool          isOutput;          /* true if signal is an output. 
                                            It can be an output too if tpt_vmapi_enableReadWrite() has been called. */

    struct tpt_vmapi_signalconfig *next; /* next config or nullptr if end-of-list */
  } tpt_vmapi_signalconfig;

  /* ========================= INTERFACE FUNCTIONS ============================= */

  /**
   * Loads the given program to the TPT-VM and returns a handle to identify it.
   *
   * @param progfile
   *     A path to a '.tvm' file, containing the program that is to be loaded.
   * @param outfile
   *     The desired path for a new '.tptbin' file, that stores the test's outputs.
   *
   * @return
   *     A handle that identifies the prepared execution and allows controlling it.
   */
  TPT_DECLARE(tpt_vmapi_handle) tpt_vmapi_createExecution(const char *progfile, const char *outfile);

  /**
   * Each TPT computation runs with a specified stepsize which is determined by the settings
   * in the "platform configuration" in TPT.
   * The function tpt_vmapi_setStepSize() allows to override this stepsize setting *after* the 
   * test programm ('.tvm' file) has been loaded (using 'tpt_vmapi_createExecution') but *before*
   * running it (using 'tpt_vmapi_runCycle').
   *
   * @param stepsize
   *      stepsize in micro seconds. Value must be greater than 0.
   */
  TPT_DECLARE(void) tpt_vmapi_setStepSize(tpt_vmapi_handle handle, tpt_uint32 stepsize);

  /**
   * Sets the value of the 'roundScalingResults' flag.
   * 
   * 1. If set, all values that are copied from TPT to SUT will be rounded up/down to the nearest integer value 
   *    if the fixtype is neiher float nor double. 
   *    Otherwise (flag is not set), all values are casted to integer.
   * 
   * 2. If set, all values that are copied from SUT to TPT will be rounded up/down to the nearest integer value 
   *    if the phystype is neither float nor double. 
   *    Otherwise (flag is not set), all values are casted to integer.
   */
  TPT_DECLARE(void) tpt_vmapi_setRoundScalingResults(tpt_vmapi_handle handle, tpt_bool enable);

  /**
   * Runs a single test program's execution cycle. The current cycle's result
   * will be stored into the output file, if one has been configured.
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   */
  TPT_DECLARE(tpt_result) tpt_vmapi_runCycle(tpt_vmapi_handle handle);

  /**
   * returns the number of test cycles that has been executed already. Initially
   * it returns 0. After each call of tpt_vmapi_runCycle() it will be incremented by 1.
   */
  TPT_DECLARE(tpt_uint32) tpt_vmapi_getCycleCount(tpt_vmapi_handle handle);

  /**
   * Alternative for 'tpt_vmapi_runCycle()'. The following sequence of calls
   * has the same effect as 'tpt_vmapi_runCycle()':
   *
   *   tpt_vmapi_internal_updateTestInputs(handle);
   *   tpt_vmapi_internal_callCycle(handle);
   *   tpt_vmapi_internal_updateTestOutputs(handle);
   *
   * These calls can be used to gather more detailled control to the api.
   */
  TPT_DECLARE(tpt_result) tpt_vmapi_internal_callCycle(tpt_vmapi_handle handle);

  /**
   * Part of the alternative, internal api. See 'tpt_vmapi_internal_callCycle()'.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_internal_updateTestInputs(tpt_vmapi_handle handle);

  /**
   * Part of the alternative, internal api. See 'tpt_vmapi_internal_callCycle()'.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_internal_updateTestOutputs(tpt_vmapi_handle handle);

  /**
   * Closes the test program. It releases external functions and memory.
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   */
  TPT_DECLARE(void) tpt_vmapi_closeExecution(tpt_vmapi_handle handle);

  /**
   * Return the dimension of a variable
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signalName
   *      Name string of the variable in TPT.
   *
   */
  TPT_DECLARE(tpt_tsize) tpt_vmapi_getDim_v2(tpt_vmapi_handle handle,
        const char *name);

  /**
   * Assigns a channel of the SUT with a given channel of TPT.
   *
   * The bind operation provides scaling and basetype conversion. The scaling rule
   * is
   *       phys = (slope * fix + bias)/(slopeInv * fix + biasInv)
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signalName
   *      Name-string of the map channel to TPT.
   * @param physBaseType
   *      Data type of the physical value (in TPT). The value "(tpt_basetype)-1" means that the type does not
   *      matter to the caller (don't care).
   * @param fixBaseType
   *      Data type of the fix point value (in SUT).
   * @param slope
   *      The slope value (factor) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param bias
   *      The bias value (offset) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param slopeInv
   *      The slopeInv value (factor) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param biasInv
   *      The biasInv value (offset) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param value
   *      Address of the value where the SUT provides the value.
   * @param dim
   *    The number of data elements in the value array, for scalars the value is 1.
   * @return
   *      Success value of the assignment.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_bindSignal_v2(tpt_vmapi_handle handle,
            const char *signalName,
            tpt_basetype physBaseType,
            tpt_basetype fixBaseType,
            tpt_double slope,
            tpt_double bias,
            tpt_double slopeInv,
            tpt_double biasInv,
            void *value,
            tpt_tsize dim);

  /**
   * Assigns a channel of the SUT with a given channel of TPT. This function is similar to tpt_vmapi_bindSignal_v2() but
   * allows to specify a function pointer to a function that dynamically delivers the current address of a variable (which
   * might change over execution time of a test case). TPT requests the address using this function (pointer) in each
   * execution cycle.
   *
   * The bind operation provides scaling and basetype conversion. The scaling rule
   * is
   *       phys = (slope * fix + bias)/(slopeInv * fix + biasInv)
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signalName
   *      Name-string of the map channel to TPT.
   * @param physBaseType
   *      Data type of the physical value (in TPT). The value "(tpt_basetype)-1" means that the type does not
   *      matter to the caller (don't care).
   * @param fixBaseType
   *      Data type of the fix point value (in SUT).
   * @param slope
   *      The slope value (factor) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param bias
   *      The bias value (offset) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param slopeInv
   *      The slopeInv value (factor) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param biasInv
   *      The biasInv value (offset) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param valueAccessFcn
   *      Function pointer of a function that delivers the current address of the value
   *      where the SUT provides the value.
   * @param dim
   *    The number of data elements in the value array, for scalars the value is 1.
   * @return
   *      Success value of the assignment.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_bindSignalDynamic_v2(tpt_vmapi_handle handle,
            const char *signalName,
            tpt_basetype physBaseType,
            tpt_basetype fixBaseType,
            tpt_double slope,
            tpt_double bias,
            tpt_double slopeInv,
            tpt_double biasInv,
            tpt_vmapi_valueaccess_fcntype_v2 valueAccessFcn,
            tpt_tsize dim);
      

   /**
   * Assigns a channel of the SUT with a given channel of TPT. This function is similar to tpt_vmapi_bindSignal_v2() but
   * allows to specify pointers to a getter function and a setter function that can retrieve and assing the value to the SUT a 
   * user-defined customizable way. 
   * TPT calls these functions in each execution cycle.
   *
   * The bind operation provides scaling and basetype conversion. The scaling rule
   * is
   *       phys = (slope * fix + bias)/(slopeInv * fix + biasInv)
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signalName
   *      Name-string of the map channel to TPT.
   * @param physBaseType
   *      Data type of the physical value (in TPT). The value "(tpt_basetype)-1" means that the type does not
   *      matter to the caller (don't care).
   * @param fixBaseType
   *      Data type of the fix point value (in SUT).
   * @param slope
   *      The slope value (factor) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param bias
   *      The bias value (offset) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param slopeInv
   *      The slopeInv value (factor) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param biasInv
   *      The biasInv value (offset) used for scaling (scaling from SUT-Fixed to TPT-Phys)
   * @param getterFcn
   *      Function pointer of a function that shall assign the current value provided by TPT to the SUT (TPT -> SUT).
   * @param setterFcn
   *      Function pointer of a function that shall retrieve the current value from the SUT (SUT -> TPT).
   * @param dim
   *    The number of data elements in the value array, for scalars the value is 1.
   * @return
   *      Success value of the assignment.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_bindSignalGetSet_v2(tpt_vmapi_handle handle,
            const char *signalName,
            tpt_basetype physBaseType,
            tpt_basetype fixBaseType,
            tpt_double slope,
            tpt_double bias,
            tpt_double slopeInv,
            tpt_double biasInv,
            tpt_vmapi_valuegetset_fcntype_v2 getterFcn,
            tpt_vmapi_valuegetset_fcntype_v2 setterFcn,
            tpt_tsize dim); 

  /**
   * Returns a list of all signals that have been bound using "tpt_vmapi_bindSignal_v2()",
   * "tpt_vmapi_bindSignalDynamic_v2()", or "tpt_vmapi_bindSignalGetSet_v2()" before. Returns TPT_NULL if this list
   * is empty; otherwise a pointer to the first item; subsequent items chained by means of the 
   * tpt_vmapi_signalconfig.next element. To iterate through the list, simply use the following pattern:
   * 
   * tpt_vmapi_signalconfig *sig = tpt_vmapi_getBoundSignals();
   * while(sig != TPT_NULL) {
   *   // ... do some computations with "sig"
   *   sig = sig->next;
   * }
   * 
   */
  TPT_DECLARE(tpt_vmapi_signalconfig*) tpt_vmapi_getBoundSignals(tpt_vmapi_handle handle);

  /**
   * Derive the type enumeration value based on a given byte size and a signed/unsigned flag. 
   * In case there is no supported type with the given byte size and signed/unsigned flag an error is reported.
   *
   * @return 
   *      The type enumeration value of type "tpt_basetype" that matches with the given byte size and signed/unsigned flag.
   */
  TPT_DECLARE(tpt_basetype) tpt_vmapi_getIntTypeEnum(const tpt_uint8 byteSize, const tpt_bool isSigned);

  /**
   * This function may be used AFTER all tpt_vmapi_bindSignal_v2() calls have been processed. 
   * Calling tpt_vmpai_bindSignalFinalize() will check whether or not *all* IN/OUT channels declared in TPT 
   * have been bound to a corresponding physical C-Code address using tpt_vmapi_bindSignal_v2() before. 
   * If not, an error will be raised (using tpt_raiseError() including a detailed error message).
   *
   * @return
   *      true if successful, false in case of an error.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_bindSignalFinalize();

  /**
   * get a handle for the given signal (based on a signal name).
   * The signal must have been bound before using tpt_vmapi_bindSignal_v2() before.
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signalName
   *      Name-string of the signal in TPT.
   * @return
   *      a handle to a signal
   */
  TPT_DECLARE(tpt_vmapi_signalhandle) tpt_vmapi_getSignal(tpt_vmapi_handle handle, const char *signalName);

  /**
   * true iff the signal has been updated during the most recent VM cycle
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signal
   *      handle of the signal in TPT (retrieved using tpt_vmapi_getSignal() before).
   * @param index
   *      index of the element to query if the signal is an array. The value of "index" must be in the
   *      range [0...dim-1]. In addition, if "index" is negative, the isUpdated() property is computed as
   *      the disjunction of the isUpdated-property of all dim elements that are part of this signal
   *      (i.e. "isUpdated(s[0]) || isUpdated(s[1]) || .. isUpdated(s[sim-1])"). In simple words: calling
   *      "tpt_vmapi_isUpdated(hVM, hSig, -1)" is true if at least one of the elements of "hSig" has been
   *      changed/updated in the most recent VM cycle.
   * @return true iff the signal has been updated during the most recent VM cycle
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_isUpdated(tpt_vmapi_handle handle, tpt_vmapi_signalhandle signal, tpt_int32 index);

  /**
   * constant representing a not-a-number. same as NAN from <math.h>
   */
  TPT_DECLARE(tpt_double) tpt_vmapi_NAN();

  /**
   * convenience macro that allows using the NAN value as a literal in source code
   */
  #define TPT_NAN (tpt_vmapi_NAN())

  /**
   * Enable/Disable read/write functionality for an output signal of TPT.
   * If read/write is enabled the output signal will be read from the SUT before each TPT cycle.
   *
   * Note that tpt_vmapi_bindSignal_v2() must be applied before calling tpt_vmapi_enableReadWrite()
   * for a signal.
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signalName
   *      Name-string of the signal in TPT.
   * @return
   *      true if successful, false in case of an error.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_enableReadWrite(tpt_vmapi_handle handle, const char *signalName);
  /**
   * If the value of an output signal of TPT should be limited (i.e. truncated/cropped) to a
   * given value range this function can be used to specify the limits. The truncation will be
   * performed while the TPT VM output value is copied to the "value" address as specified
   * in the tpt_vmapi_bindSignal_v2() function before.
   *
   * Note that tpt_vmapi_bindSignal_v2() must be applied before calling tpt_vmapi_setLimits()
   * for a signal.
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signalName
   *      Name-string of the signal in TPT.
   * @return
   *      true if successful, false in case of an error.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_setLimits(tpt_vmapi_handle handle, const char *signalName,
                                            tpt_double minLimit, tpt_double maxLimit);

  /**
   * In case of arrays the tpt_vmapi_bindSignal_v2() function just specifies the base address of the
   * first array element. In order to determine the address offset the function
   * tpt_vmapi_setIndexCoeffs_v2() can be used to specify the coefficients of the following
   * formula which will be applied to determine the address offset from the base address to the
   * address of the array element with index "idx":
   *
   * addrOffset = (idx/a)*b + (idx%a)*c
   *
   * This formula is a general purpose computation that covers a number of common use-cases of memory
   * layout explained in detail below.
   *
   * USE CASE 0: the array is a "normal" flat array without padding:
   * ===============================================================
   *   If the array in memory is "compact", i.e. there is no padding between array elements,
   *   the function tpt_vmapi_setIndexCoeffs_v2() does not have to be used in this case at all,
   *   since this is the default behavior of the TPT-VM and is semantically equivalent to the
   *   parameters
   *     a = 1
   *     b = sizeof(datatype)
   *     c = 0
   *   where datatype is the primitive datatype stored in the array. Thus, the formula can be
   *   reduced to
   *       addrOffset = idx*sizeof(datatype)
   *   which is equivalent to the natural address mapping of arrays in C language anyway.
   *   Remember: tpt_vmapi_setIndexCoeffs_v2() does not have to be used in this case, since this
   *   is the default behavior of the TPT-VM.
   *
   * USE CASE 1: the array is a flat array with padding:
   * ===================================================
   *   Padding means, that the physical layout of the array is bigger than the logical one
   *   due to memory between the array elements that does not belong to the array itself.
   *   This case usually occurs if the array is a member of an array of structs: Assume
   *   the following example structure S:
   *      struct S {
   *        int16 x;
   *        int16 y;
   *      };
   *   and an array foo as an array of struct S:
   *      // array of struct
   *      S[10] foo;
   *   Thus, the elements foo[0].x, foo[1].x, foo[2].x, ... have a padding of 2 bytes (the
   *   addresses for the struct elements "y").
   *
   *   In order to handle such cases, set "a=1" and "c=0". Thus, the formula above
   *   can be reduced to:
   *       addrOffset = idx*b
   *   "b" determines the address offset in bytes btw. the array elements. In the example
   *   above the offset must be b=4=sizeof(S).
   *
   * USE CASE 2: the array is a matrix in ROW_DIR (deposited row-by-row) without padding
   * ===================================================================================
   *   Since the TPT-VM uses row-by-row layout for matrices without padding internally,
   *   the matrix can simply be treated as a flat array (see USE CASE 0). Thus, calling
   *   tpt_vmapi_setIndexCoeffs_v2() is *not* necessary.
   *
   * USE CASE 3: the array is a matrix in ROW_DIR (deposited row-by-row) with padding
   * ==================================================================================
   *   Padding means, that the physical layout of the matrix is bigger than the logical one.
   *   Assume an uint8-matrix
   *    mx = [ 0  1  2  3
   *           4  5  6  7
   *           8  9 10 11]
   *   with dimx=4 and dimy=3. However, if the *logical* matrix dimension is just [2][2],
   *   some memory addresses are "unused":
   *    mx = [0 1 _ _
   *          2 3 _ _
   *          _ _ _ _]
   *   In other words, even if the physical layout has 3*4=12 elements, but just 2*2=4 elements
   *   are in use.
   *   In this case the meaning of the coefficients is
   *       a: logical column count (aka dimx, x-axis dimension) which must not be 0 (2 in
   *          the example above)
   *       b: number of physical bytes per row (4 in the example above, 4*1 byte per uint8)
   *       c: byte offset from one phyicial element to the next one in one row (1 in the example
   *          above, 1 byte for 1 uint8 value)
   *   In the example this results in the formula:
   *       addrOffset = (idx/2)*4 + (idx%2)*1
   *   The resulting offset sequence is addrOffset = {0, 1, 4, 5}
   *
   * USE CASE 4: the array is a matrix in COLUMN_DIR (deposited column-by-column) without padding
   * ============================================================================================
   *   The TPT-VM organizes matrices row-by-row internally. Thus, if in C code the layout is
   *   column-by-column the matrix indices must be tranposed in this case.
   *       addrOffset = (idx/a)*b + (idx%a)*c
   *   In this case the meaning of the coefficients is
   *       a: logical column count (aka: dimx, the x-axis dimension) which must not be 0
   *       b: byte offset from one phyicial element to the next one in one column
   *       c: number of physical bytes per column
   *
   * USE CASE 5: the array is a matrix in COLUMN_DIR (deposited column-by-column) with padding
   * =========================================================================================
   *   Padding means, that the physical layout of the matrix is bigger than the logical one.
   *   Assume an uint8-matrix
   *    mx = [ 0  3  6  9
   *           1  4  7 10
   *           2  5  8 11]
   *   with dimx=4 and dimy=3. However, if the *logical* matrix dimension is just [2][2],
   *   some memory addresses are "unused":
   *    mx = [0 1 _ _
   *          2 3 _ _
   *          _ _ _ _]
   *   In other words, the physical layout has 3*4=12 elements column-by-column, but the
   *   logical matrix has just 2*2=4 elements that are organized row-by-row by the TPT-VM
   *   internally.
   *   In this case the meaning of the coefficients is
   *       a: logical column count (aka dimx, x-axis dimension) which must not be 0
   *       b: byte offset from one phyicial element to the next one in one column
   *       c: number of physical bytes per column
   *   In the example this results in the formula:
   *       addrOffset = (idx/2)*1 + (idx%2)*3
   *   The resulting offset sequence is addrOffset = {0, 3, 1, 4} which are exactly the
   *   corresponding offsets in the physical layout above.
   *
   *
   * Note that tpt_vmapi_bindSignal_v2() must be applied before calling tpt_vmapi_setIndexCoeffs_v2()
   * for a signal.
   *
   * @param handle
   *      A handle to an execution session (see 'tpt_vmapi_createExecution').
   * @param signalName
   *      Name-string of the signal in TPT.
   * @param a
   *      coefficient a, see above for a detailed description of the meaning
   * @param b
   *      coefficient b, see above for a detailed description of the meaning
   * @param c
   *      coefficient c, see above for a detailed description of the meaning
   * @return
   *      true if successful, false in case of an error.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_setIndexCoeffs_v2(tpt_vmapi_handle handle,
              const char *signalName,
              tpt_uint32 a, tpt_uint32 b, tpt_uint32 c);

  /**
   * Assigns a TPT client function (i.e. a function with role CLIENT in TPT that has
   * to be implemented in the SUT) to the given implementation by means of a C function
   * pointer 'funchandler'.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_bindFunction(  tpt_vmapi_handle handle,
              const char *funcName,
              tpt_vmapi_funchandler funchandler);

  /**
   * Lookup for a function with the given ID, -1 if a function with the
   * given <code>funcName</code> does not exist. This ID can be used later to call
   * the function at runtime (within a cycle) using
   */
  TPT_DECLARE(tpt_int16) tpt_vmapi_getFunctionID(  tpt_vmapi_handle handle,
              const char *funcName);

  /**
   * Call a function with the given function ID. This ID can be determined
   * using tpt_vmapi_getFunctionID() during initialization phase of the VM
   */
  TPT_DECLARE(tpt_result) tpt_vmapi_runFunctionByID(
            tpt_vmapi_handle handle,
            tpt_uint16 funcId);

  /**
   * Call a function with the given function name. This call is just an abbrev. of
   * tpt_vmapi_getFunctionID() plus tpt_vmapi_runFunctionByID().
   */
  TPT_DECLARE(tpt_result) tpt_vmapi_runFunctionByName(
            tpt_vmapi_handle handle,
            const char *funcName);

  /**
   * Convenience function that loads a test file (infile) and generates a tptbin
   * file (outfile). The funtion pointer bindinterface should point to a function
   * (not NULL) that is used to initialized the variables. the SUT cycle must
   * contain the SUT function calls for cyclic operations.
   *
   * CAUTION: Note that implicitly a default error handler will be installed that
   * stops the process using exit() in case of any error reported by the TPT-VM.
   */
  TPT_DECLARE(void) tpt_vmapi_runTestcase(
        const char *infile,
        const char *outfile,
        tpt_vmapi_handle *handle,
        void (*bindinterface)(void),
        void (*sutcycle)(void));

  /**
   * Convenience function that loads a test file (infile) and generates a tptbin
   * file (outfile). The funtion pointer bindinterface should point to a function
   * (not NULL) that is used to initialized the variables. the SUT cycle must
   * contain the SUT function calls for cyclic operations.
   * In contrast to tpt_vmapi_runTestcase() the file path
   * arguments are wide character strings encoded as UTF-16.
   *
   * CAUTION: Note that implicitly a default error handler will be installed that
   * stops the process using exit() in case of any error reported by the TPT-VM.
   */
  TPT_DECLARE(void) tpt_vmapi_runTestcase_v2(
        const tpt_wchar *infile,
        const tpt_wchar *outfile,
        tpt_vmapi_handle *handle,
        void (*bindinterface)(void),
        void (*sutcycle)(void));
 
  /**
   * Returns the value of 'TPT_TRUE', if an error occurred during test runtime, otherwise 'TPT_FALSE'.
   *
   * @return
   *      The value 'TPT_TRUE' of type 'tpt_bool', if an error occurred during test runtime, otherwise 'TPT_FALSE'.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_hasError();

  /**
   * Returns the code of the last error that occurred within the API.
   * @Deprecated Use tpt_vmapi_getErrorStatus() to get a more precise status code.
   *
   * @return
   *      A value of enum 'tpt_vmapi_errcode' which is the code of the last error that occurred within the API.
   *
   */
  TPT_DECLARE(tpt_uint16) tpt_vmapi_getErrorCode();

  /**
   * Returns the error message, the test execution provides, if an error occurred during test runtime.
   *
   * @return
   *      The error message, that the test execution provides, if an error occurred during test runtime.
   */
  TPT_DECLARE(const char *) tpt_vmapi_getErrorMessage();

  /**
   * Clears the last error that occurred within the API.
   */
  TPT_DECLARE(void) tpt_vmapi_resetError();

  /**
   * Exchange of parameter values in the SuT to TPT
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_exchangeParamsFromSuT(
        const char *inFileName,
        const char *outFileName,
        tpt_vmapi_handle *handle,
        void (*bindinterface)(void));
        
  /**
   * Exchange of parameter values in the SuT to TPT
   * In contrast to tpt_vmapi_exchangeParamsFromSuT() the file path 
   * arguments are wide character strings encoded as UTF-16.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_exchangeParamsFromSuT_v2(
        const tpt_wchar *inFileName,
        const tpt_wchar *outFileName,
        tpt_vmapi_handle *handle,
        void (*bindinterface)(void));
  
  /**
   * Exchange of parameter values from TPT to the SuT
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_exchangeParamsToSuT(
        const char *inFileName,
        tpt_vmapi_handle *handle,
        void (*bindinterface)(void));
    
  /**
   * Exchange of parameter values from TPT to the SuT
   * In contrast to tpt_vmapi_exchangeParamsToSuT() the file path 
   * argument is a wide character string encoded as UTF-16.
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_exchangeParamsToSuT_v2(
        const tpt_wchar *inFileName,
        tpt_vmapi_handle *handle,
        void (*bindinterface)(void));
    
  /**
   * Installs realtime-command, in case the simulation time is determined by the SuT.
   * Default for realtime-command is TPT_NULL. In case of that, it is assumed
   * that realtime is equal to simulation time, which means a multiple of the step size.
   */
  TPT_DECLARE(void) tpt_vmapi_installRealtimeCmd(tpt_vmapi_realtimeCommand rtCmd);

  /* =================================================================================================================
   * =================================================================================================================
   * +++ DEPRECATED FUNCTIONS: +++
   * =================================================================================================================
   * ================================================================================================================= */


  /**
   * +++ Deprecated since TPT16 +++
   * 
   * this function should not be used any longer since array size is 32bit now (uint32). Use tpt_vmapi_getDim_v2() instead. 
   */
  TPT_DECLARE(tpt_uint16) tpt_vmapi_getDim(tpt_vmapi_handle handle,
        const char *name);


  /**
   * +++ Deprecated since TPT16 +++
   * 
   * this function should not be used any longer since array size is 32bit now (uint32). Use tpt_vmapi_bindSignal_v2() instead. 
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_bindSignal(tpt_vmapi_handle handle,
            const char *signalName,
            tpt_basetype physBaseType,
            tpt_basetype fixBaseType,
            tpt_double slope,
            tpt_double bias,
            tpt_double slopeInv,
            tpt_double biasInv,
            void *value,
            tpt_uint16 dim);        

  /**
   * +++ Deprecated since TPT16 +++
   * 
   * this type should not be used any longer since array size is 32bit now (uint32). Use tpt_vmapi_valueaccess_fcntype_v2 instead. 
   */
  typedef void* (*tpt_vmapi_valueaccess_fcntype)(tpt_uint16 idx, tpt_uint8 dims_size, const tpt_uint16 *dims);


  /**
   * +++ Deprecated since TPT16 +++
   * 
   * this function should not be used any longer since array size is 32bit now (uint32). Use tpt_vmapi_bindSignalDynamic_v2() instead. 
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_bindSignalDynamic(tpt_vmapi_handle handle,
            const char *signalName,
            tpt_basetype physBaseType,
            tpt_basetype fixBaseType,
            tpt_double slope,
            tpt_double bias,
            tpt_double slopeInv,
            tpt_double biasInv,
            tpt_vmapi_valueaccess_fcntype valueAccessFcn,
            tpt_uint16 dim);

  /**
   * +++ Deprecated since TPT16 +++
   * 
   * this type should not be used any longer since array size is 32bit now (uint32). Use tpt_vmapi_valuegetset_fcntype_v2 instead. 
   */
  typedef tpt_bool (*tpt_vmapi_valuegetset_fcntype)(void* valPtr, tpt_uint16 idx, tpt_uint8 dims_size, const tpt_uint16 *dims);

  /**
   * +++ Deprecated since TPT16 +++
   * 
   * this function should not be used any longer since array size is 32bit now (uint32). Use tpt_vmapi_bindSignalGetSet_v2() instead. 
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_bindSignalGetSet(tpt_vmapi_handle handle,
            const char *signalName,
            tpt_basetype physBaseType,
            tpt_basetype fixBaseType,
            tpt_double slope,
            tpt_double bias,
            tpt_double slopeInv,
            tpt_double biasInv,
            tpt_vmapi_valuegetset_fcntype getterFcn,
            tpt_vmapi_valuegetset_fcntype setterFcn,
            tpt_uint16 dim); 

  /**
   * +++ Deprecated since TPT17 +++
   * 
   * this function should not be used any longer since the coefficients are 32bit now (uint32). Use tpt_vmapi_setIndexCoeffs_v2() instead. 
   */
  TPT_DECLARE(tpt_bool) tpt_vmapi_setIndexCoeffs(tpt_vmapi_handle handle,
              const char *signalName,
              tpt_int16 a, tpt_int16 b, tpt_int16 c);

#ifdef __cplusplus
}
#endif

#endif /* __TPT_VM_API_H_ */
