/*********************************************************************
 *              Copyright (c) 2021 PikeTec GmbH
 *               CONFIDENTIAL AND PROPRIETARY.
 *                   ALL RIGHTS RESERVED
 *              - PROPRIETARY OF PikeTec GmbH -
 **********************************************************************/
/* clang-format off */

#ifndef __TPT_FUSION_BASE_H_
#define __TPT_FUSION_BASE_H_

#ifdef __cplusplus
extern "C"
{
#endif

/**
 * Handle to a fusion.
 */
typedef int tpt_fusion_handle;

/**
 * Handle to a fusion node.
 */
typedef int tpt_fusion_node_handle;

/**
 * Handle to a variable.
 */
typedef int tpt_fusion_variable_handle;

/**
 * Handle to a function.
 */
typedef int tpt_fusion_function_handle;

/**
 * Handle to an asynchronous job.
 */
typedef int tpt_fusion_job_handle;

/**
 * Function type for a "job" that can be executed asynchronously
 * 
 * @param data 
 *     pointer to a data segment which has been specified as the second argument in tpt_fusion_runAsync() 
 *     before (can be used for context structs provided to the job function)
 */
typedef void (*tpt_fusion_job)(void *data);

/**
 * Function type of fusion_function handlers
 * @param node
 *     handle of the node that implements the function (the one that called tpt_fusion_functionRegister()
 * @param fhandle
 *     handle of the function itself
 * @param args
 *     pointer that has been specified at tpt_fusion_functionRegister(). This pointer can be used by
 *     implementers to mimic function closures.
 */
typedef void (*tpt_fusion_function_handler)(tpt_fusion_node_handle node, 
                                            tpt_fusion_function_handle fhandle, 
                                            void *args);

/**
 * The enum of variable types.
 */
typedef enum _tpt_fusion_variable_type_enum {
  TPT_FUSION_VARIABLE_TYPE_SCALAR,
  TPT_FUSION_VARIABLE_TYPE_STRING,
  TPT_FUSION_VARIABLE_TYPE_ARRAY,
  TPT_FUSION_VARIABLE_TYPE_TABLE,
  TPT_FUSION_VARIABLE_TYPE_CURVE,
  TPT_FUSION_VARIABLE_TYPE_MAP
} tpt_fusion_variable_type;
  
/**
 * The enum of variable roles.
 */
typedef enum _tpt_fusion_variable_role_enum {
  TPT_FUSION_VARIABLE_ROLE_SIGNAL,
  TPT_FUSION_VARIABLE_ROLE_PARAMETER,
  TPT_FUSION_VARIABLE_ROLE_MEASUREMENT
} tpt_fusion_variable_role;

/**
 * This is a handler function type that is used for requesting the value of variables that are 
 * provided "lazily", i.e. the value is only retrieved for a simulation cycle if the value 
 * is requested for reading using one of the functions 
 *    tpt_fusion_readValue(),
 *    tpt_fusion_readValueAsLong(),
 *    tpt_fusion_readValues(),
 *    tpt_fusion_readValuesAsLong(),
 *    tpt_fusion_isWrittenAsLong()
 *
 * The handler function is then responsible to write the corresponding value to the fusion (using
 * one of the functions above) *before* the handler function terminates.
 */
typedef void (*tpt_fusion_lazy_handler)(tpt_fusion_node_handle handle, 
                                        tpt_fusion_variable_handle vhandle);

/* ===================================== FOR BACKWARD COMPATIBILTY =========================== */

/**
 * Handle to a parameter.
 */
typedef tpt_fusion_variable_handle tpt_fusion_signal_handle;

/**
 * Handle to a parameter.
 */
typedef tpt_fusion_variable_handle tpt_fusion_param_handle;

/* =========================================================================================== */

/**
 * Fusion error codes.
 *
 * See 'tpt_fusion_getErrorCode()'.
 */  
typedef enum _tpt_fusion_errcode {
  TPT_FUSION_ERRCODE_NO_ERROR             = 0, 
  TPT_FUSION_ERRCODE_IO_ERROR             = 100, 
  TPT_FUSION_ERRCODE_MEM_ERROR            = 200, 
  TPT_FUSION_ERRCODE_ABORT                = 300,
  TPT_FUSION_ERRCODE_NODE_ERROR_STARTUP   = 401,
  TPT_FUSION_ERRCODE_NODE_ERROR_INIT      = 402,
  TPT_FUSION_ERRCODE_NODE_ERROR_CALLFCN   = 403,
  TPT_FUSION_ERRCODE_NODE_ERROR_SHUTDOWN  = 404,
  TPT_FUSION_ERRCODE_INTERNAL_ERROR       = 900
} tpt_fusion_errcode;

#ifndef tpt_long
  #ifdef __GNUC__
    #define tpt_long long long
  #else
    #define tpt_long __int64
  #endif
#endif

#ifdef __cplusplus
}
#endif

/*
 * The following macros are relevant to properly define the import/export directives in case of a DLL creation or a DLL access.
 */
#define TPT_DECLARE_EXPORT(x) __declspec(dllexport) x
#define TPT_DECLARE_IMPORT(x) __declspec(dllimport) x

#ifdef TPT_FUSION_EXPORTS
  #define TPT_DECLARE_FUSION(x) TPT_DECLARE_EXPORT(x)
#else
  #define TPT_DECLARE_FUSION(x) TPT_DECLARE_IMPORT(x)
#endif

#ifdef TPT_FUSION_NODE_EXPORTS
  #define TPT_DECLARE_FUSION_NODE(x) TPT_DECLARE_EXPORT(x)
#else
  #define TPT_DECLARE_FUSION_NODE(x) TPT_DECLARE_IMPORT(x)
#endif

#ifdef VERSIONINFO_EXPORTS
  #define TPT_DECLARE_FUSION_NODE_VERSIONINFO(x) TPT_DECLARE_EXPORT(x)
#else
  #define TPT_DECLARE_FUSION_NODE_VERSIONINFO(x) TPT_DECLARE_IMPORT(x)
#endif

#endif /* __TPT_FUSION_BASE_H_ */
