/* File    : fusion_mrun.c
 *
 * Contains several helper functions for fusion_main in order to access and run the Simulink model.
 *
 * Compiler specified defines:
 *  RT              - Required.
 *  MODEL=modelname - Required.
 *  NUMST=#         - Required. Number of sample times.
 *	NCSTATES=#      - Required. Number of continuous states.
 *  TID01EQ=1 or 0  - Optional. Only define to 1 if sample time task
 *                    id's 0 and 1 have equal rates.
 *  MULTITASKING    - Optional. (use MT for a synonym).
 *	SAVEFILE        - Optional (non-quoted) name of .mat file to create.
 *			  		  Default is <MODEL>.mat
 *
 */
 
/* ============================ INCLUDES =========================== */
 
#include <float.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "rtw_capi.h"
#include "rtmodel.h"
#include "rt_sim.h"
#include "rt_logging.h"
#ifdef UseMMIDataLogging
#include "rt_logging_mmi.h"
#endif

#include "ext_work.h"

#include "tpt_fusion.h"
 
/* ============================ DEFINES =========================== */
 
#ifndef TRUE
#define FALSE (0)
#define TRUE  (1)
#endif

#ifndef EXIT_FAILURE
#define EXIT_FAILURE  1
#endif
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS  0
#endif

#define QUOTE1(name) #name
#define QUOTE(name) QUOTE1(name)    /* need to expand name    */

#ifndef RT
# error "must define RT"
#endif

#ifndef MODEL
# error "must define MODEL"
#endif

#ifndef NUMST
# error "must define number of sample times, NUMST"
#endif

#ifndef NCSTATES
# error "must define NCSTATES"
#endif

#ifndef SAVEFILE
# define MATFILE2(file) #file ".mat"
# define MATFILE1(file) MATFILE2(file)
# define MATFILE MATFILE1(MODEL)
#else
# define MATFILE QUOTE(SAVEFILE)
#endif

#define RUN_FOREVER -1.0

#define EXPAND_CONCAT(name1,name2) name1 ## name2
#define CONCAT(name1,name2) EXPAND_CONCAT(name1,name2)
#define RT_MODEL            CONCAT(MODEL,_rtModel)

#ifdef EXT_MODE
#  define rtExtModeSingleTaskUpload(S)                          \
   {                                                            \
        int stIdx;                                              \
        rtExtModeUploadCheckTrigger(rtmGetNumSampleTimes(S));   \
        for (stIdx=0; stIdx<NUMST; stIdx++) {                   \
            if (rtmIsSampleHit(S, stIdx, 0 /*unused*/)) {       \
                rtExtModeUpload(stIdx,rtmGetTaskTime(S,stIdx)); \
            }                                                   \
        }                                                       \
   }
#else
#  define rtExtModeSingleTaskUpload(S) /* Do nothing */
#endif

#define MAINCAPIFCN CONCAT(MODEL,_GetCAPIStaticMap)

#define STR_EXPAND(tok) #tok
#define STR(tok) STR_EXPAND(tok)

/* ====================== EXTERNAL FUNCTIONS ====================== */

#ifdef __cplusplus

extern "C" {

#endif

extern RT_MODEL *MODEL(void);
extern void MdlInitializeSizes(void);
extern void MdlInitializeSampleTimes(void);
extern void MdlStart(void);
extern void MdlOutputs(int_T tid);
extern void MdlUpdate(int_T tid);
extern void MdlTerminate(void);
extern const rtwCAPI_ModelMappingStaticInfo* MAINCAPIFCN();

#ifdef __cplusplus

}
#endif

#if NCSTATES > 0
#ifdef __cplusplus

extern "C" {

#endif
  extern void rt_ODECreateIntegrationData(RTWSolverInfo *si);
  extern void rt_ODEUpdateContinuousStates(RTWSolverInfo *si);
#ifdef __cplusplus

}
#endif

# define rt_CreateIntegrationData(S) \
    rt_ODECreateIntegrationData(rtmGetRTWSolverInfo(S));
# define rt_UpdateContinuousStates(S) \
    rt_ODEUpdateContinuousStates(rtmGetRTWSolverInfo(S));
# else
# define rt_CreateIntegrationData(S)  \
      rtsiSetSolverName(rtmGetRTWSolverInfo(S),"FixedStepDiscrete");
# define rt_UpdateContinuousStates(S) /* Do Nothing */
#endif

/* ========================= GLOBAL DATA ========================== */

typedef enum _tpt_fusion_node_result_enum {
  TPT_FUSION_NODE_RESULT_OK,
  TPT_FUSION_NODE_RESULT_ERROR,
  TPT_FUSION_NODE_RESULT_FINISHED
  
} tpt_fusion_node_result;

static struct {
  int_T    stopExecutionFlag;
  int_T    isrOverrun;
  int_T    overrunFlags[NUMST];
  int_T    eventFlags[NUMST];
  const    char_T *errmsg;
} GBLbuf;

const real_T finaltime = RUN_FOREVER;
RT_MODEL *S;

const rtwCAPI_ModelMappingStaticInfo* tpt_fusion_simulink_getCAPIStaticMap();
void** tpt_fusion_simulink_getDataAddressMap();
tpt_fusion_node_result tpt_fusion_simulink_initModel(tpt_fusion_node_handle handle, double fusStSz);
tpt_fusion_node_result tpt_fusion_simulink_terminateModel(tpt_fusion_node_handle handle);
tpt_fusion_node_result tpt_fusion_simulink_rtOneStep(tpt_fusion_node_handle handle);
char* tpt_fusion_simulink_modelName();
static void tpt_fusion_simulink_applyFusionSampleTime(double fusStSz);

/* ====================== FUNCTIONS ======================= */

const rtwCAPI_ModelMappingStaticInfo* tpt_fusion_simulink_getCAPIStaticMap() {
	return MAINCAPIFCN();
} /* end tpt_fusion_simulink_getCAPIStaticMap */

void** tpt_fusion_simulink_getDataAddressMap() {

	rtwCAPI_ModelMappingInfo mmiTmp = S->DataMapInfo.mmi;
	
	return mmiTmp.InstanceMap.dataAddrMap;
	
} /* end tpt_fusion_simulink_getDataAddressMap */

char* tpt_fusion_simulink_modelName() {
	
	return STR(MODEL);
	
} /* end tpt_fusion_simulink_modelName */

tpt_fusion_node_result tpt_fusion_simulink_initModel(tpt_fusion_node_handle handle, double fusStSz) {
	
	const char *status;

	/* initialize global memory for model execution */
	(void)memset(&GBLbuf, 0, sizeof(GBLbuf));
	
    /* initialize the model */
    
    S = MODEL();
	if (rtmGetErrorStatus(S) != NULL) {
		tpt_fusion_printf(handle, "error during model registration: %s\n", rtmGetErrorStatus(S));
        return TPT_FUSION_NODE_RESULT_ERROR;
    }
	rtmSetTFinal(S,finaltime);
	MdlInitializeSizes();
	
	/* sample time settings */
    MdlInitializeSampleTimes();
	tpt_fusion_simulink_applyFusionSampleTime(fusStSz);
	status = rt_SimInitTimingEngine(rtmGetNumSampleTimes(S),
                                    rtmGetStepSize(S),
                                    rtmGetSampleTimePtr(S),
                                    rtmGetOffsetTimePtr(S),
                                    rtmGetSampleHitPtr(S),
                                    rtmGetSampleTimeTaskIDPtr(S),
                                    rtmGetTStart(S),
                                    &rtmGetSimTimeStep(S),
                                    &rtmGetTimingData(S));
	if (status != NULL) {
        tpt_fusion_printf(handle, "failed to initialize sample time engine: %s\n", status);
        return TPT_FUSION_NODE_RESULT_ERROR;
    }
	
	/* further model initialization steps */
	rt_CreateIntegrationData(S);
#ifdef UseMMIDataLogging
    rt_FillStateSigInfoFromMMI(rtmGetRTWLogInfo(S),&rtmGetErrorStatus(S));
#endif
    GBLbuf.errmsg = rt_StartDataLogging(rtmGetRTWLogInfo(S),
                                        rtmGetTFinal(S),
                                        rtmGetStepSize(S),
                                        &rtmGetErrorStatus(S));
    if (GBLbuf.errmsg != NULL) {
        tpt_fusion_printf(handle, "error starting data logging: %s\n", GBLbuf.errmsg);
        return TPT_FUSION_NODE_RESULT_ERROR;
    }
    rtExtModeCheckInit(rtmGetNumSampleTimes(S));
    rtExtModeWaitForStartPkt(rtmGetRTWExtModeInfo(S),
                             rtmGetNumSampleTimes(S),
                             (boolean_T *)&rtmGetStopRequested(S));
    MdlStart();
    if (rtmGetErrorStatus(S) != NULL) {
      GBLbuf.stopExecutionFlag = 1;
    }
	
	return TPT_FUSION_NODE_RESULT_OK;
	
} /* end tpt_fusion_simulink_initModel */

tpt_fusion_node_result tpt_fusion_simulink_terminateModel(tpt_fusion_node_handle handle) {
	
#ifdef UseMMIDataLogging
	rt_CleanUpForStateLogWithMMI(rtmGetRTWLogInfo(S));
#endif
	rt_StopDataLogging(MATFILE,rtmGetRTWLogInfo(S));
	
	rtExtModeShutdown(rtmGetNumSampleTimes(S));
	
#ifdef MULTITASKING
	if (!GBLbuf.isrOverrun) {
		int_T i;
		for (i=1; i<NUMST; i++) {
			if (GBLbuf.overrunFlags[i]) {
				tpt_fusion_printf(handle, "%s ISR overrun - sampling rate is too fast for sample time index %d\n", QUOTE(MODEL), i);
				return TPT_FUSION_NODE_RESULT_ERROR;
			}
		}
	}
#endif

	MdlTerminate();
	
	return TPT_FUSION_NODE_RESULT_OK;
	
} /* end tpt_fusion_simulink_terminateModel */

static void tpt_fusion_simulink_applyFusionSampleTime(double fusStSz) {

	double fStepSize = fusStSz / 1000000;
	S->Timing.sampleTimes[0] = (fStepSize);
	S->Timing.stepSize = (fStepSize);
	S->Timing.stepSize0 = (fStepSize);
	rtsiSetFixedStepSize(&S->solverInfo, fStepSize);
	
} /* end tpt_fusion_simulink_applyFusionSampleTime */

#if !defined(MULTITASKING)  /* SINGLETASKING */

tpt_fusion_node_result tpt_fusion_simulink_rtOneStep(tpt_fusion_node_handle handle) {

    real_T tnext;

	if (GBLbuf.stopExecutionFlag || rtmGetStopRequested(S))
		return TPT_FUSION_NODE_RESULT_FINISHED;
		
		/* external mode */
    rtExtModeOneStep(rtmGetRTWExtModeInfo(S),
                rtmGetNumSampleTimes(S),
                (boolean_T *)&rtmGetStopRequested(S));
	
    /***********************************************
     * Check and see if base step time is too fast *
     ***********************************************/
    if (GBLbuf.isrOverrun++) {
        GBLbuf.stopExecutionFlag = 1;
		tpt_fusion_printf(handle, "ISR overrun - base sampling rate is too fast\n");
        return TPT_FUSION_NODE_RESULT_ERROR;
    }

    /***********************************************
     * Check and see if error status has been set  *
     ***********************************************/
    if (rtmGetErrorStatus(S) != NULL) {
        GBLbuf.stopExecutionFlag = 1;
        tpt_fusion_printf(handle, "ErrorStatus set: \"%s\"\n", rtmGetErrorStatus(S));
        return TPT_FUSION_NODE_RESULT_ERROR;
    }

    /* enable interrupts here */

    tnext = rt_SimGetNextSampleHit();
    rtsiSetSolverStopTime(rtmGetRTWSolverInfo(S),tnext);

    MdlOutputs(0);

    rtExtModeSingleTaskUpload(S);

    GBLbuf.errmsg = rt_UpdateTXYLogVars(rtmGetRTWLogInfo(S),
                                        rtmGetTPtr(S));
    if (GBLbuf.errmsg != NULL) {
        GBLbuf.stopExecutionFlag = 1;
        tpt_fusion_printf(handle, "%s\n", GBLbuf.errmsg);
        return TPT_FUSION_NODE_RESULT_ERROR;
    }

    MdlUpdate(0);
    rt_SimUpdateDiscreteTaskSampleHits(rtmGetNumSampleTimes(S),
                                       rtmGetTimingData(S),
                                       rtmGetSampleHitPtr(S),
                                       rtmGetTPtr(S));

    if (rtmGetSampleTime(S,0) == CONTINUOUS_SAMPLE_TIME) {
        rt_UpdateContinuousStates(S);
    }

    GBLbuf.isrOverrun--;

    rtExtModeCheckEndTrigger();
	
	return TPT_FUSION_NODE_RESULT_OK;
	
} /* end tpt_fusion_simulink_rtOneStep */

#else /* MULTITASKING */

# if TID01EQ == 1
#  define FIRST_TID 1
# else
#  define FIRST_TID 0
# endif

tpt_fusion_node_result tpt_fusion_simulink_rtOneStep(tpt_fusion_node_handle handle) {

    int_T  i;
    real_T tnext;
    int_T  *sampleHit = rtmGetSampleHitPtr(S);

	if (GBLbuf.stopExecutionFlag || rtmGetStopRequested(S))
		return TPT_FUSION_NODE_RESULT_FINISHED;
	
	  /* external mode */
    rtExtModeOneStep(rtmGetRTWExtModeInfo(S),
                rtmGetNumSampleTimes(S),
                (boolean_T *)&rtmGetStopRequested(S));
	
    /***********************************************
     * Check and see if base step time is too fast *
     ***********************************************/
    if (GBLbuf.isrOverrun++) {
        GBLbuf.stopExecutionFlag = 1;
        tpt_fusion_printf(handle, "ISR overrun - base sampling rate is too fast\n");
        return TPT_FUSION_NODE_RESULT_ERROR;
    }

    /***********************************************
     * Check and see if error status has been set  *
     ***********************************************/
    if (rtmGetErrorStatus(S) != NULL) {
        GBLbuf.stopExecutionFlag = 1;
        tpt_fusion_printf(handle, "ErrorStatus set: \"%s\"\n", rtmGetErrorStatus(S));
        return TPT_FUSION_NODE_RESULT_ERROR;
    }
    /* enable interrupts here */

    /***********************************************
     * Update discrete events                      *
     ***********************************************/
    tnext = rt_SimUpdateDiscreteEvents(rtmGetNumSampleTimes(S),
                                       rtmGetTimingData(S),
                                       rtmGetSampleHitPtr(S),
                                       rtmGetPerTaskSampleHitsPtr(S));
    rtsiSetSolverStopTime(rtmGetRTWSolverInfo(S),tnext);
    for (i=FIRST_TID+1; i < NUMST; i++) {
        if (sampleHit[i] && GBLbuf.eventFlags[i]++) {
            GBLbuf.isrOverrun--; 
            GBLbuf.overrunFlags[i]++;    /* Are we sampling too fast for */
            GBLbuf.stopExecutionFlag=1;  /*   sample time "i"?           */
            return TPT_FUSION_NODE_RESULT_ERROR;
        }
    }
    /*******************************************
     * Step the model for the base sample time *
     *******************************************/
    MdlOutputs(FIRST_TID);

    rtExtModeUploadCheckTrigger(rtmGetNumSampleTimes(S));
    rtExtModeUpload(FIRST_TID,rtmGetTaskTime(S, FIRST_TID));

    GBLbuf.errmsg = rt_UpdateTXYLogVars(rtmGetRTWLogInfo(S),
                                        rtmGetTPtr(S));
    if (GBLbuf.errmsg != NULL) {
        GBLbuf.stopExecutionFlag = 1;
        tpt_fusion_printf(handle, "%s\n", GBLbuf.errmsg);
        return TPT_FUSION_NODE_RESULT_ERROR;
    }

    MdlUpdate(FIRST_TID);

    if (rtmGetSampleTime(S,0) == CONTINUOUS_SAMPLE_TIME) {
        rt_UpdateContinuousStates(S);
    }
     else {
        rt_SimUpdateDiscreteTaskTime(rtmGetTPtr(S), 
                                     rtmGetTimingData(S), 0);
    }

#if FIRST_TID == 1
    rt_SimUpdateDiscreteTaskTime(rtmGetTPtr(S), 
                                 rtmGetTimingData(S),1);
#endif

    /************************************************************************
     * Model step complete for base sample time, now it is okay to          *
     * re-interrupt this ISR.                                               *
     ************************************************************************/

    GBLbuf.isrOverrun--;

    /*********************************************
     * Step the model for any other sample times *
     *********************************************/
    for (i=FIRST_TID+1; i<NUMST; i++) {
        /* If task "i" is running, don't run any lower priority task */
        if (GBLbuf.overrunFlags[i]) return TPT_FUSION_NODE_RESULT_OK;

        if (GBLbuf.eventFlags[i]) {
            GBLbuf.overrunFlags[i]++;

            MdlOutputs(i);
 
            rtExtModeUpload(i, rtmGetTaskTime(S,i));

            MdlUpdate(i);

            rt_SimUpdateDiscreteTaskTime(rtmGetTPtr(S), 
                                         rtmGetTimingData(S),i);

            /* Indicate task complete for sample time "i" */
            GBLbuf.overrunFlags[i]--;
            GBLbuf.eventFlags[i]--;
        }
    }

    rtExtModeCheckEndTrigger();
	
	return TPT_FUSION_NODE_RESULT_OK;
	
} /* end tpt_fusion_simulink_rtOneStep */

#endif /* MULTITASKING */

/* EOF: fusion_mrun.c */
 