package com.etas.ascet.api.v1.datatype;

import org.apache.commons.lang3.StringUtils;
import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;

import com.google.auto.value.AutoValue;
import com.ryanharter.auto.value.gson.GenerateTypeAdapter;

/**
 * Array dimension
 */
// Checkstyle:off
@NonNullByDefault
@GenerateTypeAdapter
@AutoValue
public abstract class ArrayDimensionDtoV1 {

    /**
     * The count of indexable positions in one dimension direction.
     * When the size is a system constant that is resolved at compile time,
     * the size is the maximum value of the sysconst and the symbol name is set.
     * <p/>
     * Either none, the size or the size+sizeSymbol are set.<br/>
     * If none is set, it is an unconstrained array and the size is given in its instantiation.
     */
    @Nullable
    public abstract Integer size();

    /**
     * When the size is a system constant that is resolved at compile time,
     * the size is the maximum value of the sysconst and the symbol name is the name of the sysconst.
     * Either none, the size or the size+sizeSymbol are set.<br/>
     * If none is set, it is an unconstrained array and the size is given in its instantiation.
     */
    @Nullable
    public abstract String sizeSymbol();

    /**
     * Normally this is zero.
     * But in ESDL an array can have a dimension starting from another value than zero. Example:
     * <p/>
     *
     * <pre>
     * type myindex is integer 2 .. 4;
     * type myarray is array[myindex] of integer;
     * </pre>
     *
     * In ESDL {@code myarray} can only be indexed with 2 .. 4. 0 is an invalid index.<br/>
     * In the generated C-Code, the array is a standard C array with starting index 0.
     * The example array {@code myarray} from above would have a valid index range 0 .. 3.
     */
    public abstract int startIndex();

    /**
     * Normally this is integer.
     * But e.g. a derived type or an enum type might be used.
     */
    public abstract IdentifierDtoV1 indexType();

    public static Builder builder() {
        return new AutoValue_ArrayDimensionDtoV1.Builder()
                .sizeSymbol(StringUtils.EMPTY);
    }

    @AutoValue.Builder
    public interface Builder {

        /**
         * @see ArrayDimensionDtoV1#size()
         */
        Builder size(@Nullable Integer value);

        /**
         * @see ArrayDimensionDtoV1#sizeSymbol()
         */
        Builder sizeSymbol(@Nullable String value);

        /**
         * @see ArrayDimensionDtoV1#startIndex()
         */
        Builder startIndex(int value);

        /**
         * @see ArrayDimensionDtoV1#indexType()
         */
        Builder indexType(IdentifierDtoV1 value);

        ArrayDimensionDtoV1 build();
    }
}
