package com.etas.ascet.api.v1.datatype;

import java.util.List;

import org.eclipse.jdt.annotation.NonNullByDefault;
import org.eclipse.jdt.annotation.Nullable;

import com.etas.ascet.api.KeyValueDto;
import com.google.auto.value.AutoValue;
import com.ryanharter.auto.value.gson.GenerateTypeAdapter;

/**
 * Represents an instance as it is generated for the target (vs. model). This can be e.g. a scalar class field variable
 * (leaf of instance tree), or an instantiation of a class/structure within the instance tree. Note, that also top level
 * singleton classes, static classes/fields and data-interface-messages are part of the instance tree. The containing
 * data-interface itself on the other hand is not part of the instance tree.
 */
// Checkstyle:off
@NonNullByDefault
@GenerateTypeAdapter
@AutoValue
public abstract class TargetInstanceDtoV1 {

    /**
     * Identifier that is independent of code generation settings.
     * This identifier can be used to find matching elements in runs with different settings.
     * package_sub_Type.field.field
     */
    public abstract IdentifierDtoV1 identifier();

    /**
     * The link to the containing {@link TargetInstanceDtoV1}, e.g. a class or struct,
     * that declares this item as a field. Or empty, when this is a root item,
     * or if there is no containing instance (static fields).
     */
    @Nullable
    public abstract IdentifierDtoV1 containingInstance();

    /**
     * The link to the type of the containingInstance, e.g. a class, struct type or data-interface,
     * that declares this item as a field. Or empty, when this is declared at package level.
     */
    @Nullable
    public abstract IdentifierDtoV1 containerType();

    /**
     * Instance path with or without the package of the top level element,
     * depending on the code gen setting legacy mode.
     * For static elements are contained in the surrounding type name.
     * Public static elements and messages simple name only.
     *
     * <pre>
     * package_sub_Type.field.field
     * Type.field.field
     *
     * class A {
     *     static real fld1;
     * }
     * -> [package.]A.fld1
     *
     * class A {
     *     public static real fld1;
     *     provides message real fld2;
     * }
     * -> fld1
     * -> fld2
     *
     * data interface DI {
     *     real fld1;
     * }
     * -> fld1
     * </pre>
     */
    public abstract String modelPath();

    /**
     * The link to the type of this instance.
     */
    public abstract IdentifierDtoV1 type();

    /**
     * Only applicable for instances, which cannot have child instances
     */
    @Nullable
    public abstract TargetInstanceValueDtoV1 value();

    /**
     * the semantic comment if existing or empty string
     */
    public abstract String comment();

    /**
     * the type of declaration, e.g. characteristic vs variable vs constant ...
     */
    public abstract DeclKindV1 declKind();

    /**
     * public or private
     */
    public abstract VisibilityV1 visibility();

    public abstract boolean isStatic();

    public abstract boolean isVirtual();

    /**
     * is it a provide declaration (provide/require)
     */
    public abstract boolean isProvide();

    /**
     * Path to C variable/field. Is empty string for classes.
     * For constants and system constants this value is an empty string.
     */
    public abstract String generatedAccessPath();

    /**
     * Empty list, or in case of array/matrix/table the information about the dimensions
     */
    public abstract List<TargetInstanceDimensionDtoV1> dimensions();

    /**
     * The value of {@code @a2l_unit_label} annotation
     */
    public abstract String a2lUnit();

    /**
     * The direction of the message or an empty value. Scalar/Enum/Structs can be used as message.
     */
    @Nullable
    public abstract MessageDirectionV1 messageDirection();

    public static final String PROPERTY_KEY_INTERPOLATION = "interpolation";
    public static final String PROPERTY_KEY_ASAP2_COEFFICIENTS = "asap-coefficients";
    public static final String PROPERTY_KEY_IS_FIXED_TABLE = "is-fixed-table";
    public static final String PROPERTY_KEY_IS_CONTAINED_IN_CLASS_ACCESSING_MSG =
            "is-contained-in-class-accessing-message";

    /**
     * Other properties, that are not part of the API. They might be enhanced during program evolution.
     */
    public abstract List<KeyValueDto> properties();

    public static Builder builder() {
        return new AutoValue_TargetInstanceDtoV1.Builder();
    }

    @AutoValue.Builder
    public interface Builder {
        /**
         * @see TargetInstanceDtoV1#identifier()
         */
        Builder identifier(IdentifierDtoV1 value);

        /**
         * @see TargetInstanceDtoV1#containingInstance()
         */
        Builder containingInstance(@Nullable IdentifierDtoV1 value);

        /**
         * @see TargetInstanceDtoV1#containerType()
         */
        Builder containerType(@Nullable IdentifierDtoV1 value);

        /**
         * @see TargetInstanceDtoV1#modelPath()
         */
        Builder modelPath(String value);

        /**
         * @see TargetInstanceDtoV1#type()
         */
        Builder type(IdentifierDtoV1 value);

        /**
         * @see TargetInstanceDtoV1#value()
         */
        Builder value(@Nullable TargetInstanceValueDtoV1 value);

        /**
         * @see TargetInstanceDtoV1#comment()
         */
        Builder comment(String value);

        /**
         * @see TargetInstanceDtoV1#declKind()
         */
        Builder declKind(DeclKindV1 value);

        /**
         * @see TargetInstanceDtoV1#generatedAccessPath()
         */
        Builder generatedAccessPath(String value);

        /**
         * @see TargetInstanceDtoV1#visibility()
         */
        Builder visibility(VisibilityV1 value);

        /**
         * @see TargetInstanceDtoV1#isStatic()
         */
        Builder isStatic(boolean value);

        /**
         * @see TargetInstanceDtoV1#isVirtual()
         */
        Builder isVirtual(boolean value);

        /**
         * @see TargetInstanceDtoV1#isProvide()
         */
        Builder isProvide(boolean value);

        /**
         * @see TargetInstanceDtoV1#dimensions()
         */
        Builder dimensions(List<TargetInstanceDimensionDtoV1> value);

        /**
         * @see TargetInstanceDtoV1#a2lUnit()
         */
        Builder a2lUnit(String value);

        /**
         * @see TargetInstanceDtoV1#messageDirection()
         */
        Builder messageDirection(@Nullable MessageDirectionV1 value);

        /**
         * @see TargetInstanceDtoV1#properties()
         */
        Builder properties(List<KeyValueDto> value);

        TargetInstanceDtoV1 build();
    }
}
