/* RegistryKey.java
 *
 * jRegistryKey - A JNI wrapper of the Windows Registry functions.
 * Copyright (c) 2001 - 2004 BEQ Technologies Inc.
 * 9407 - 27 Avenue
 * Edmonton, Alberta
 * T6N - 1L9 Canada
 * (780) 430-0056
 * (780) 437-6121 (fax)
 * http://www.beq.ca
 *
 * Original Author: Joe Robinson <joe@beq.ca>
 * Current Maintainer: Matt Hilliard <matt@beq.ca>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package ca.beq.util.win32.registry;

import java.util.Iterator;

/**
 * A representation of system registry keys, <code>RegistryKey</code> is the
 * principle class of the <code>ca.beq.util.win32.registry</code> package.
 * 
 * <p>
 * A <code>RegistryKey</code> is composed of two components (specified at
 * creation):
 * 
 * <ol>
 * <li>A <i>root key</i>, and
 * <li>A <i>path</i>
 * </ol>
 * <p>
 * A <code>RegistryKey</code> may also begin with a machine componenet, in
 * (formatted as a Windows UNC (This means 4 backslash '\' characters from Java)
 * and will connect to the given machine subject to suitable permission on the
 * remote machine.
 * 
 * <p>
 * The root key component is defined to be any valid <code>RootKey</code>
 * object. The path component is similar to a file-system pathname; it is a
 * series of zero or more string <i>names</i> separated by backslahes (This
 * means 2 backslash '\' characters from Java). Unlike file-system pathnames
 * there is no "optional prefix", and each name defines a registry key. For
 * example, <i>Software\\BEQ Technologies\\Test</i> is a valid registry key,
 * whereas <i>\\Software\\BEQ Technologies\\Test</i> is not.
 * 
 * <p>
 * The following code snippet demonstrates creating, reading, and deleting
 * values in the registry using <code>RegistryKey</code>:
 * 
 * <pre>
 * // initialise the libraray to a specific DLL
 * RegistryKey.initialize(pathToDll);
 * 
 * // create a new key, &quot;Test&quot;, under HKLM
 * RegistryKey r = new RegistryKey(RootKey.HKEY_LOCAL_MACHINE, &quot;Test&quot;);
 * if (!r.exists()) {
 * 	r.create();
 * } // if
 * 
 * // create value entries
 * RegistryValue v = new RegistryValue(&quot;aString&quot;, ValueType.REG_SZ, &quot;test&quot;);
 * r.setValue(v);
 * 
 * v.setName(&quot;aDword&quot;);
 * v.setType(ValueType.REG_DWORD);
 * v.setData(new Integer(0x1001001));
 * r.setValue(v);
 * 
 * // read value entries
 * Iterator i = r.values();
 * while (i.hasNext()) {
 * 	v = (RegistryValue) i.next();
 * 	System.out.println(v.toString());
 * } // while
 * 
 * // delete registry key
 * r.delete();
 * </pre>
 * 
 * @author BEQ Technologies Inc.
 * @version $Revision: 1.13 $
 * 
 * <pre>
 * $Id: RegistryKey.java,v 1.13 2006/07/10 02:17:48 bayblade Exp $
 * 
 * $Log: RegistryKey.java,v $
 * Revision 1.13  2006/07/10 02:17:48  bayblade
 * Updated javadocs to reflect out 1.4 code changes
 * 
 * Revision 1.12  2006/07/10 01:41:11  bayblade
 * Remote Registry and last access time goodness, courtesy Dave Kipfer &lt;kipfer@cae.com&gt;
 * 
 * Revision 1.11  2005/12/07 05:06:10  bayblade
 * Reordered code
 * 
 * Revision 1.10  2005/12/05 03:22:59  bayblade
 * Hacked to accomodate the versioning sync between the jni and the java portions of the code.  Note that a version mismatch on implicit calls will spew data on stderr, and a mismatch on an explicit call will throw a Registry exception, although initialization will have occured anyhow.  As this is error, not common behavior, its not expected to break anything that's not broken anyhow.
 * 
 * Revision 1.9  2005/01/27 19:34:08  bayblade
 * Cleaned up base constructor which sets values to m_path and m_root, despite these values already being set before the constructor.
 * 
 * Revision 1.8  2004/12/20 21:21:30  bayblade
 * Revampled initialization process so 1.0 upgraders weren't so confused;  fixed bug with IBM machines; and accidentally reverted to ANSI formatting instead of K&amp;R in a few spots
 * 
 * Revision 1.7  2004/04/20 00:16:18  bayblade
 * Added &lt;pre&gt; tags in headers so change history log 
 * can be sensibly read from javadoc
 * 
 * Revision 1.6  2004/04/13 19:17:06  mollinie
 * Reviewed the initialization code. In particular, it has to be invoked explicitely
 * by user code (in order to have control of _where_ the initialization will fail, should
 * it fail).
 * Added methods to test whether initialization succeeded.
 * Call to constructors of the class will fail unless initialization succeeded (in order
 * to enforce the initialize-before-use policy).
 * 
 * Revision 1.5  2004/04/07 23:55:45  mollinie
 * Corrected a corner-case bug in accessing a subkey from the root key.
 * 
 * Revision 1.4  2004/04/07 18:57:21  mollinie
 * Renamed member fields according to our conventions.
 * 
 * Revision 1.3  2004/03/27 01:01:01  mollinie
 * Preliminary reformating.
 * 
 * Revision 1.2  2004/03/27 00:38:19  mollinie
 * Preliminary reformating.
 * Added a proposal for enhancement.
 * 
 * Revision 1.1.1.1  2004/03/25 22:05:38  bayblade
 * Initial Import
 * </pre>
 */
public class RegistryKey {

	/** Indicates whether library initialization succeeded. */
	// package
	static boolean c_initSucceeded = false;

	/** library that gets loaded if no name is presented */
	public static final String DEFAULT_LIB = "jRegistryKey";

	/** we seem to use it alot, make it a constant */
	private static final String EMPTY = "";

	/** ENV variable potentially pointing to native dll */
	public static final String ENVIRONMENT_PATH = "JREGISTRYKEY_LIB";

	/** CLI variable potentially pointing to native dll */
	public static final String PROPERTY_PATH = "jRegistryKey.lib";

	/** constant splitter for registry path heirarchy */
	private static final char SPLITTER = '\\';

	/**
	 * Checks whether the native library was initialized successfully.
	 * 
	 * @deprecated use
	 * @link #isInitialized() and throw IllegalStateException if false.
	 *       <p>
	 *       As of release 1.2, this no longer throws IllegalStateException when
	 *       the library was not loaded successfully. It remains only for
	 *       signature compatability with previous releases.
	 */
	public static void checkInitialized() {
	};

	private static void compareVersions() {
		String versionAPI = getAPIVersion();
		if (versionAPI == null) {
			return;
		}
		String versionJNI;
		try {
			versionJNI = getJNIVersion();
		} catch (UnsatisfiedLinkError e) {
			versionJNI = "pre 1.3.2";
		}

		if (versionAPI.equals(versionJNI)) {
			return;
		}
		throw new RegistryException(
				"Native library loaded, however JNI version " + versionJNI
						+ " does not match API version " + versionAPI);
	}

	/**
	 * Exposes the build revistion
	 * 
	 * @return string mapping to the binay build. If unset at runtime, this
	 *         value is usually "undefined"
	 */
	public static String getAPIVersion() {
		return Version.current;
	}

	private static native String getJNIVersion();

	/**
	 * Called to implicitly initialize the native library for jRegistryKey. By
	 * cascading through names potentially stored in the System property
	 * "jRegistryKey.lib", then the System environment variable
	 * "JREGISTRYKEY_LIB" and finally opting to test the default system name
	 * "jRegistryKey"
	 * 
	 * This method may be called to explicitly before calling any other code in
	 * the jRegistryKey library as can the "explicit call" to
	 * 
	 * @link #initialize(String) however should neither be called explicitly
	 *       prior to library access, this method will be called implicitly.
	 */
	public static void initialize() {
		// do nothing if we chain up a large number of implicit calls to this
		// method
		if (isInitialized()) {
			return;
		}

		try {
			// make sure it isn't loaded and throw UnsatisfiedLinkError
			testInitialized();
			return;

		} catch (UnsatisfiedLinkError ule) {
			// resume processing because we didn't return
		}

		// try and use property to load our DLL
		String property = System.getProperty(PROPERTY_PATH);
		if (loadNative(property)) {
			return;
		}

		// try and use environment to load our DLL
		try {
			String env = System.getenv(ENVIRONMENT_PATH);
			if (loadNative(env)) {
				return;
			}

		} catch (Error depricated) {
			// JREs 1.3 and 1.4 will, while 1.5 won't--just process as usual
		}

		// default to passed library name to load our DLL
		if (loadNative(DEFAULT_LIB)) {
			return;
		}

		// run out of loading attempts, call this to throw UnsatisfiedLinkError
		testInitialized();

	}

	/**
	 * Explicitly initializes the native library for jRegistryKey named
	 * <code>libname</code>. This method may be called to explicitly use a
	 * jRegistryKey system library or be passed a jRegistryKey library as a
	 * complete path. This method may be called to explicitly before calling any
	 * other code in the jRegistryKey library as can the "implicit call" to
	 * 
	 * @link #initialize() however should neither be called explicitly prior to
	 *       library access, the implicit call will be made and this method will
	 *       do nothing.
	 * 
	 * @param libname
	 *            The named system native library or complete filename and path
	 *            to the jRegistryKey native library to use.
	 */
	public synchronized static void initialize(String libname) {
		// do nothing if we chain up a large number of implicit calls to this
		// method
		if (isInitialized()) {
			return;
		}

		try {
			// make sure it isn't loaded and throw UnsatisfiedLinkError
			testInitialized();
			return;

		} catch (UnsatisfiedLinkError ule) {
			// resume processing because we didn't return
		}

		// default to passed library name to load our DLL
		if (loadNative(libname)) {
			return;
		}

		// run out of loading attempts, call this to throw UnsatisfiedLinkError
		testInitialized();
	}

	/**
	 * Tests whether the native library was initialized successfully.
	 * 
	 * @return <code>true</code> if the library was loaded successfully, else
	 *         <code>false</code>.
	 */
	public synchronized static boolean isInitialized() {
		return c_initSucceeded;
	}

	/**
	 * Attempt to load the native library as a system dll, and failing that try
	 * again as a complete pathname.
	 * 
	 * @param library
	 *            name or path of our native dll
	 * @return true if one of the loads succeeded, false if neither load
	 *         succeeded
	 */
	private static boolean loadNative(String library) {

		if (library == null) {
			return false;
		}

		try {
			System.loadLibrary(library);
			c_initSucceeded = true;

		} catch (UnsatisfiedLinkError uleName) {
			try {
				System.load(library);
				c_initSucceeded = true;

			} catch (UnsatisfiedLinkError ulePath) {
			}

		}

		if (c_initSucceeded) {
			compareVersions();
		}
		return c_initSucceeded;
	}

	/**
	 * Tests whether library initialization succeeded. When calling it, either
	 * the call will quietly do nothing--and the library was initialized
	 * successfully--or the call results in an UnsatisfiedLinkError being
	 * thrown--in which case the library was _not_ initialized successfully.
	 */
	// package
	static native void testInitialized();

	// package
	String m_path = EMPTY;

	// package
	RootKey m_root = RootKey.HKEY_CURRENT_USER;

	// package
	String m_Machine = null;

	// package
	boolean m_ignore_expand = false;

	/**
	 * Constructs a new <code>RegistryKey</code> referencing the root path of
	 * the <code>RootKey.HKEY_CURRENT_USER</code> root key.
	 */
	public RegistryKey() {
		// implicit initialization. trust JITs to remove this hereafter.
		try {
			RegistryKey.initialize();
		} catch (RegistryException re) {
			re.printStackTrace();
		}
	}

	/**
	 * Constructs a new <code>RegistryKey</code> referencing the root path of
	 * the specified <code>RootKey</code>.
	 * 
	 * @param root
	 *            the root key of this <code>RegistryKey</code>
	 */
	public RegistryKey(RootKey root) {
		this();
		m_root = root;
	}

	/**
	 * Constructs a new <code>RegistryKey</code> referencing the specified
	 * path of the specified <code>RootKey</code>.
	 * 
	 * @param root
	 *            the root key of this <code>RegistryKey</code>
	 * @param path
	 *            the path of this <code>RegistryKey</code>
	 */
	public RegistryKey(RootKey root, String path) {
		m_root = root;
		m_path = path;
	}

	/**
	 * Constructs a new <code>RegistryKey</code> referencing the specified
	 * path of the <code>RootKey.HKEY_CURRENT_USER</code> root key.
	 * 
	 * @param path
	 *            the path of this <code>RegistryKey</code>
	 */
	public RegistryKey(String path) {
		this();
		m_path = path;
	}

	/**
	 * Constructs a new <code>RegistryKey</code> referencing the root path of
	 * the specified <code>RootKey</code> on machine <code>iMachine</code>.
	 * 
	 * @param iMachine
	 *            the name of the remote machine, formatted <i>\\computername</i>.
	 * @param root
	 *            the root key of this <code>RegistryKey</code>
	 */
	public RegistryKey(String iMachine, RootKey root) {
		this(root);
		m_Machine = iMachine;
	}

	/**
	 * Constructs a new <code>RegistryKey</code> referencing the specified
	 * path of the specified <code>RootKey</code> on machine
	 * <code>iMachine</code>.
	 * 
	 * @param iMachine
	 *            the name of the remote machine, formatted <i>\\computername</i>.
	 * @param root
	 *            the root key of this <code>RegistryKey</code>
	 * @param path
	 *            the path of this <code>RegistryKey</code>
	 */
	public RegistryKey(String iMachine, RootKey root, String path) {
		this(root, path);
		m_Machine = iMachine;
	}

	/**
	 * Creates this registry key in the system registry.
	 * 
	 * @throws RegistryException
	 *             if this registry key already exists in the registry
	 */
	public native void create();

	/**
	 * Creates the specified subkey.
	 * 
	 * @param name
	 *            name the subkey to be created
	 * 
	 * @return the created subkey
	 * 
	 * @throws RegistryException
	 *             if this registry key does not exist in the registry, or if
	 *             the specified subkey name already exists.
	 */
	public RegistryKey createSubkey(String name) {
		RegistryKey r = new RegistryKey(m_root, makePath(name));
		r.create();
		return r;
	}

	/**
	 * Deletes this registry key and any subkeys or values from the system
	 * registry.
	 * 
	 * <p>
	 * <b>WARNING: This method can potentially cause catastrophic damage to the
	 * system registry. USE WITH EXTREME CARE!</b>
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public native void delete();

	/**
	 * Deletes the specified value from this registry key.
	 * 
	 * @param name
	 *            the name of the value to be deleted
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry,
	 *             or if the specified value is not possess by this registry key
	 */
	public native void deleteValue(String name);

	/**
	 * Tests if this registry key, as defined by the <code>RootKey</code> and
	 * <code>path</code>, currently exists in the system registry.
	 * 
	 * @return <code>true</code> if this key exists in the registry
	 */
	public native boolean exists();

	/**
	 * Returns this <code>RegistryKey</code>'s <code>name</code>. The name
	 * is last portion of a <code>path</code>. For example, for the
	 * <code>path</code> <i>Software\\BEQ Technologies\\TestKey</i>, the
	 * <code>name</code> would be <i>TestKey</i>.
	 * 
	 * @return the <code>name</code>
	 */
	public String getName() {
		return m_path.substring(1 + m_path.lastIndexOf(SPLITTER));
	}

	/**
	 * Returns this <code>RegistryKey</code>'s <code>path</code>.
	 * 
	 * @return the <code>path</code>
	 */
	public String getPath() {
		return m_path;
	}

	/**
	 * Returns this <code>RegistryKey</code>'s root key.
	 * 
	 * @return the root key
	 */
	public RootKey getRootKey() {
		return m_root;
	}

	/**
	 * Returns a <code>long</code> representing a Windows FILETIME of the last
	 * instance the registry was modified. <strong>Note that Windows 95, 98 and
	 * ME return 0 for this value</strong>
	 * 
	 * @return the last modified time of the registry key in 100ns since January
	 *         1, 1601 (UTC)
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	protected native long getLastModifiedFileTime();

	/**
	 * Returns a <code>long</code> representing the time since the epoch of
	 * the last instance the registry was modified. This is a suitable value for
	 * comparing with System.currentTimeMillis() or wrapping in a java.util.Date
	 * object. <strong>Note that Windows 95, 98 and ME return 0 for this value</strong>
	 * 
	 * @return the last modified time of the registry key in ms since January 1,
	 *         1970 (UTC)
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public long getLastModifiedTime() {
		// this started with an estimate and was improved by newton's method
		final long MS_BETWEEN_IDIOCY_AND_EPOCH = 11644473600000L;
		return Math.round(getLastModifiedFileTime() * 0.0001)
				- MS_BETWEEN_IDIOCY_AND_EPOCH;
	}

	/**
	 * Returns a <code>RegistryValue</code> representing the specified value.
	 * 
	 * @param name
	 *            the name of the value to be retreived
	 * 
	 * @return the RegistryValue of the specified value name
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry,
	 *             or if this registry key does not possess the specified value
	 */
	public native RegistryValue getValue(String name);

	/**
	 * Tests if this registry key possesses the specified subkey.
	 * 
	 * @param name
	 *            the subkey to test for
	 * 
	 * @return <code>true</code> if this registry key possesses subkeys
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public boolean hasSubkey(String name) {
		RegistryKey r = new RegistryKey(m_root, makePath(name));
		return r.exists();
	}

	/**
	 * Tests if this registry key possesses subkeys. Use <code>subkeys</code>
	 * to retrieve an iterator for available subkeys.
	 * 
	 * @return <code>true</code> if this registry key possesses subkeys
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public native boolean hasSubkeys();

	/**
	 * Tests if this registry key possess the specified value.
	 * 
	 * @param name
	 *            the name of the value to be tested
	 * 
	 * @return <code>true</code> if this registry keys possess the specified
	 *         value
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public native boolean hasValue(String name);

	/**
	 * Tests if this registry key possess any values.
	 * 
	 * @return <code>true</code> if this registry keys possess any values
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public native boolean hasValues();

	/**
	 * Builds a subpath based on this path and the <code>subPath</code>
	 * argument. Takes into account the corner case where the path is empty
	 * (equivalent to this key representing a root key).
	 */
	public String makePath(String subPath) {
		StringBuffer path = new StringBuffer(m_path);
		if (!EMPTY.equals(m_path)) {
			path.append(SPLITTER);
		}
		path.append(subPath);
		return path.toString();
	}

	/**
	 * Sets the properties of a registry value according to the properties of
	 * the specified <code>RegistryValue</code>. If the specified value
	 * exists, it will be modified; if not, it will be created.
	 * 
	 * @param value
	 *            the <code>RegistryValue</code>
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public native void setValue(RegistryValue value);

	/**
	 * Returns an iterator for available subkeys.
	 * 
	 * @return an iterator of <code>RegistryKey</code>'s
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public Iterator subkeys() {
		return new KeyIterator(this);
	}

	/**
	 * Returns a string representation of this <code>RegistryKey</code>.
	 * 
	 * @return a string representation of this <code>RegistryKey</code>
	 */
	public String toString() {
		StringBuffer toString = new StringBuffer(m_root.toString());
		if (!EMPTY.equals(m_path)) {
			toString.append(SPLITTER);
			toString.append(m_path);
		}
		return toString.toString();
	}

	/**
	 * Returns an iterator for available values.
	 * 
	 * @return an iterator of <code>RegistryValue</code>'s
	 * 
	 * @throws RegistryException
	 *             if this registry key does not already exist in the registry
	 */
	public Iterator values() {
		return new ValueIterator(this);
	}

	/**
	 * The default behavior when reading REG_EXPAND_SZ keys is to expand the
	 * keys, replacing values with environment variables as required. Setting
	 * this to true changes the behavior to return the value as though it was a
	 * regular REG_SZ.
	 * 
	 * @param doReplace
     */
	public void bypassExpansion(boolean doBypass) {
		m_ignore_expand = doBypass;
	}
}