/* ValueType.java
 *
 * jRegistryKey - A JNI wrapper of the Windows Registry functions.
 * Copyright (c) 2001, BEQ Technologies Inc.
 * 9407 - 27 Avenue
 * Edmonton, Alberta
 * T6N - 1L9 Canada
 * (780) 430-0056
 * (780) 437-6121 (fax)
 * http://www.beq.ca
 *
 * Original Author: Joe Robinson <joe@beq.ca>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ca.beq.util.win32.registry;

import java.io.Serializable;

/**
 * Registry values (data) can be stored in various formats (types), represented
 * by <code>ValueType</code>. <code>ValueType</code> exposes these types in
 * a <a
 * href="http://developer.java.sun.com/developer/Books/shiftintojava/page1.html#replaceenums">typesafe
 * enum</a> that applications may use to indicate or query the type of data
 * stored in a <code>RegistryValue</code>
 * 
 * @author BEQ Technologies Inc.
 * @version $Revision: 1.12 $
 * 
 * <pre>
 * $Id: ValueType.java,v 1.12 2006/07/10 02:04:25 bayblade Exp $
 * 
 * $Log: ValueType.java,v $
 * Revision 1.12  2006/07/10 02:04:25  bayblade
 * Added some missing doccumentation for untested ValueTypes.  Just because its never been used doesn't mean it *might* not work.
 * 
 * Revision 1.11  2005/12/05 03:22:59  bayblade
 * Hacked to accomodate the versioning sync between the jni and the java portions of the code.  Note that a version mismatch on implicit calls will spew data on stderr, and a mismatch on an explicit call will throw a Registry exception, although initialization will have occured anyhow.  As this is error, not common behavior, its not expected to break anything that's not broken anyhow.
 * 
 * Revision 1.10  2005/09/30 00:27:03  bayblade
 * Unbroke confusion wrt BIG and LITTLE endian storage.
 *  We now remain consistent with what comes out in regedit.
 * 
 * Revision 1.9  2005/05/10 01:19:15  bayblade
 * Clarified poor documentation about the dword types.
 * 
 * Revision 1.8  2005/05/10 01:01:48  bayblade
 * 3 things happened here: added support for reg_qword, support for reg_multi_sz and finally unhooped unicode character conversion.
 * 
 * Revision 1.7  2005/05/04 21:20:15  bayblade
 * Added REG_QWORD, fixed BIG and LITTLE endian DWORDS
 * 
 * Revision 1.6  2004/12/20 21:21:31  bayblade
 * Revampled initialization process so 1.0 upgraders weren't so confused;  fixed bug with IBM machines; and accidentally reverted to ANSI formatting instead of K&amp;R in a few spots
 * 
 * Revision 1.5  2004/04/20 00:16:18  bayblade
 * Added &lt;pre&gt; tags in headers so change history log 
 * can be sensibly read from javadoc
 * 
 * Revision 1.4  2004/04/07 18:59:13  mollinie
 * Renamed member fields according to our conventions.
 * 
 * Revision 1.3  2004/04/07 18:24:55  mollinie
 * Added poor man's support for extra registry data types, including REG_MULTI_SZ, REG_LINK, REG_FULL_RESOURCE_DESCRIPTOR and REG_RESOURCE_REQUIREMENTS_LIST.
 * 
 * Revision 1.2  2004/03/27 01:01:01  mollinie
 * Preliminary reformating.
 * 
 * Revision 1.1.1.1  2004/03/25 22:05:38  bayblade
 * Initial Import
 * </pre>
 */

public class ValueType implements Serializable {

	/*
	 * This may be hypersensitive to versioning changes, VM changes, phases of
	 * the moon and all kinds of verbal abuse, but hey--its Serialization, its
	 * not like using Serialization hasn't earned you some pain (it has, btw)
	 */
	private static final long serialVersionUID = (long) ValueType.class
			.hashCode()
			* Version.current.hashCode();

	// implicit initialization. see method for details.
	static {
		try {
			RegistryKey.initialize();
		} catch (RegistryException re) {
			re.printStackTrace();
		}
	}

	/** The value type name displayed by <code>toString</code> */
	final String m_name;

	/**
	 * The value type integer value used by the native Windows registry
	 * functions
	 */
	final int m_value;

	/**
	 * Constructs a new value type with the specified display name and value.
	 * 
	 * @param name -
	 *            the display name of the value type
	 * @param value -
	 *            the integer value of the value type
	 */
	ValueType(String name, int value) {
		m_name = name;
		m_value = value;
	}

	/**
	 * Returns the integer value (used by the native Windows registry functions)
	 * of the value type.
	 * 
	 * @return the integer value of the value type.
	 */
	protected int getValue() {
		return m_value;
	}

	/**
	 * Returns the display name of the value type.
	 * 
	 * @return a string representation of the value type.
	 */
	public String toString() {
		return m_name;
	}

	/**
	 * The <code>REG_NONE</code> data type represents data with no defined
	 * type.
	 */
	public static final ValueType REG_NONE = new ValueType("REG_NONE", 0);

	/**
	 * The <code>REG_SZ</code> data type represents a null-terminated string.
	 */
	public static final ValueType REG_SZ = new ValueType("REG_SZ", 1);

	/**
	 * The <code>REG_EXPAND_SZ</code> data type represents a null-terminated
	 * string that contains unexpanded references to environment variables (for
	 * example, "%PATH%").
	 */
	public static final ValueType REG_EXPAND_SZ = new ValueType(
			"REG_EXPAND_SZ", 2);

	/**
	 * The <code>REG_BINARY</code> data type represents binary data in any
	 * form.
	 */
	public static final ValueType REG_BINARY = new ValueType("REG_BINARY", 3);

	/**
	 * The <code>REG_DWORD</code> data type represents a 32-bit number Despite
	 * Java being big-endian, we stick this to LITTLE_ENDIAN to maintain
	 * consistency.
	 */
	public static final ValueType REG_DWORD = new ValueType("REG_DWORD", 4);

	/**
	 * The <code>REG_DWORD_BIG_ENDIAN</code> data type represents a 32-bit
	 * number in little-endian format.
	 * 
	 * In big-endian format, a multi-byte value is stored in memory from the
	 * highest byte (the "big end") to the lowest byte. For example, the value
	 * 0x12345678 is stored as (0x12 0x34 0x56 0x78) in big-endian format.
	 */
	public static final ValueType REG_DWORD_BIG_ENDIAN = new ValueType(
			"REG_DWORD_BIG_ENDIAN", 5);

	/**
	 * The <code>REG_DWORD_LITTLE_ENDIAN</code> data type represents a 32-bit
	 * number in little-endian format (in jRK, this is equivalent to
	 * <code>REG_DWORD</code>)..
	 * 
	 * <p>
	 * In little-endian format, a multi-byte value is stored in memory from the
	 * lowest byte (the "little end") to the highest byte. For example, the
	 * value 0x12345678 is stored as (0x78 0x56 0x34 0x12) in little-endian
	 * format.
	 */
	public static final ValueType REG_DWORD_LITTLE_ENDIAN = new ValueType(
			"REG_DWORD_LITTLE_ENDIAN", 4);

	/**
	 * The <code>REG_LINK</code> data type is intended to act as a recistry
	 * link between two redistry keys, for example, mapping a "current" version
	 * to a specific one, or providing a shortcut from one location in the
	 * registry to another. <br>
	 * The link should be passed as a Unicode (utf-16) string where the named
	 * value becomes a reference to the String passed. <br>
	 * This key is untested by the author.
	 */
	public static final ValueType REG_LINK = new ValueType("REG_LINK", 6);

	/**
	 * The <code>REG_MULTI_SZ</code> data type represents an array of
	 * null-terminated strings, terminated by two null characters.
	 * 
	 * <p>
	 * The strings are stored as unicode (utf-16) in the windows registry and
	 * are null-terminated. An additional null string terminates the multi.
	 */
	public static final ValueType REG_MULTI_SZ = new ValueType("REG_MULTI_SZ",
			7);

	/**
	 * The <code>REG_RESOURCE_LIST</code> data type represents a device-driver
	 * resource list withing the resource map tree. <br>
	 * Any access vai this key will need to be done via byte[]. <br>
	 * This key is untested by the author.
	 */
	public static final ValueType REG_RESOURCE_LIST = new ValueType(
			"REG_RESOURCE_LIST", 8);

	/**
	 * The <code>REG_FULL_RESOURCE_DESCRIPTOR</code> data type represents a
	 * device-driver resource list withing the resource descriptor tree. <br>
	 * Any access vai this key will need to be done via byte[]. <br>
	 * This key is untested by the author.
	 */
	public static final ValueType REG_FULL_RESOURCE_DESCRIPTOR = new ValueType(
			"REG_FULL_RESOURCE_DESCRIPTOR", 9);

	/**
	 * The <code>REG_RESOURCE_REQUIREMENTS_LIST</code> data type represents a
	 * device-driver requirement list withing the resource descriptor tree. <br>
	 * Any access vai this key will need to be done via byte[]. <br>
	 * This key is untested by the author.
	 */
	public static final ValueType REG_RESOURCE_REQUIREMENTS_LIST = new ValueType(
			"REG_RESOURCE_REQUIREMENTS_LIST", 10);

	/**
	 * The <code>REG_QWORD</code> data type represents a 64-bit number.
	 */
	public static final ValueType REG_QWORD = new ValueType("REG_QWORD", 11);

} // fin
