/* RegistryKeyTest.java
 *
 * jRegistryKey - A JNI wrapper of the Windows Registry functions.
 * Copyright (c) 2001, BEQ Technologies Inc.
 * 9407 - 27 Avenue
 * Edmonton, Alberta
 * T6N - 1L9 Canada
 * (780) 430-0056
 * (780) 437-6121 (fax)
 * http://www.beq.ca
 *
 * Original Author: Joe Robinson <joe@beq.ca>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ca.beq.util.win32.registry;

import java.util.Arrays;
import java.util.Iterator;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * <code>RegistryKeyTest</code> tests the <code>RegistryKey</code> class.
 *
 * @see ca.beq.util.win32.registry.RegistryKey
 *
 * @author BEQ Technologies Inc.
 * @version $Revision: 1.6 $
 * 
 * $Id: RegistryKeyTest.java,v 1.6 2005/12/05 03:23:37 bayblade Exp $
 * 
 * $Log: RegistryKeyTest.java,v $
 * Revision 1.6  2005/12/05 03:23:37  bayblade
 * Beautified
 *
 * Revision 1.5  2005/05/10 01:01:48  bayblade
 * 3 things happened here: added support for reg_qword, support for reg_multi_sz and finally unhooped unicode character conversion.
 *
 * Revision 1.4  2004/04/13 19:18:55  mollinie
 * Added initialization call prior to test run.
 *
 * Revision 1.3  2004/04/02 18:24:54  mollinie
 * Corrected an assertion where the expected and actual fields where inverted.
 *
 * Revision 1.2  2004/03/27 01:01:02  mollinie
 * Preliminary reformating.
 *
 * Revision 1.1.1.1  2004/03/25 22:05:38  bayblade
 * Initial Import
 *
 */
public class RegistryKeyTest extends TestCase {

    private static final String TESTKEY = "jRegistryKeyTestKey";

    /**
     * Executable entry point.
     */
    public static void main( String[] args ) {
        junit.textui.TestRunner.run( suite() );
    }

    /**
     * Constructor.
     */
    public RegistryKeyTest( String name ) {
        super( name );
    }

    /**
     * Returns list of available tests.
     */
    public static Test suite() {
        return new TestSuite( RegistryKeyTest.class );
    }

    /**
     * Tests that the <code>exists()</code> method succeeds on existing key
     */
    public void testExistsValid() {
        RegistryKey.initialize();

        RegistryKey r = new RegistryKey( RootKey.HKEY_LOCAL_MACHINE, "HARDWARE" );
        assertEquals(
                "Valid key \"HKLM\\HARDWARE\" must report true (exists)!",
                true, r.exists() );
    }

    /**
     * Tests that the <code>exists()</code> method fails on a non-existant key
     */
    public void testExistsInvalid() {
        RegistryKey.initialize();

        RegistryKey r = new RegistryKey( RootKey.HKEY_LOCAL_MACHINE,
                "GoodBadImTheOneWithTheGun" );
        assertEquals(
                "Bogus key \"HKCU\\OogaBooga\" must report false (does not exist)!",
                false, r.exists() );
    }

    /**
     * Tests creating a non-existant key succeeds
     */
    public void testCreateNew() {
        RegistryKey.initialize();

        try {
            RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
            if ( r.exists() ) {
                r.delete();
            }

            r.create();
        } catch ( RegistryException re ) {
            fail( "Creation of a new key must succeed!" );
        }
    }

    /**
     * Tests creating a duplicate key fails
     */
    public void testCreateDuplicate() {
        RegistryKey.initialize();

        try {
            RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
            if ( !r.exists() ) {
                r.create();
            }

            r.create();
            fail( "Creation of a duplicate key must fail!" );
        } catch ( RegistryException re ) {
            // success
        }
    }

    /**
     * Tests that deleting a valid key succeeds
     */
    public void testDeleteValid() {
        RegistryKey.initialize();

        try {
            RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
            if ( !r.exists() ) {
                r.create();
            }

            r.delete();
        } catch ( RegistryException re ) {
            fail( "Deletion of a non-empty key must succeed!" );
        }
    }

    /**
     * Tests that deleting an invalid key fails
     */
    public void testDeleteInvalid() {
        RegistryKey.initialize();

        try {
            RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
            if ( r.exists() ) {
                r.delete();
            }

            r.delete();
            fail( "Deletion of an invalid key must fail!" );
        } catch ( RegistryException re ) {
            // success
        }
    }

    /**
     * Tests that the <code>hasSubkeys()</code> method fails on a key without
     * subkeys
     */
    public void testHasSubkeysFail() {
        RegistryKey.initialize();

        RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );

        try {
            if ( r.exists() ) {
                r.delete();
            }

            r.create();
            assertEquals(
                    "Checking if a key without subkey has keys must fail!",
                    false, r.hasSubkeys() );
        } catch ( RegistryException re ) {
            fail( "Subkey creation must succeed!" );
        }
    }

    /**
     * Tests that the <code>hasSubkeys()</code> method succeeds on a key that
     * has subkeys
     */
    public void testHasSubkeys() {
        RegistryKey.initialize();

        RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );

        try {
            if ( r.exists() ) {
                r.delete();
            }

            r.create();
            r.createSubkey( TESTKEY );
            assertEquals(
                    "Checking if a key with subkey has keys must not fail!",
                    true, r.hasSubkeys() );
        } catch ( RegistryException re ) {
            fail( "Subkey creation must succeed!" );
        }
    }

    static String REGVALUE_SZ = "This is a test with odd characters:\\Startmen";

    /**
     * Test that the <code>setValue()</code> and <code>getValue()</code> methods
     * succeed.
     */
    public void testSetGetValue() {
        RegistryKey.initialize();

        String REGVALUE_EXPAND_SZ = "PATH=%path%";
        Integer REGVALUE_DWORD = new Integer( 1 );
        Integer REGVALUE_DWORD_LITTLE_ENDIAN = new Integer( 2 );
        Long REGVALUE_QWORD = new Long( -1 );
        byte[] REGVALUE_BINARY = new byte[] { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };

        // test REG_SZ
        RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
        RegistryValue v = new RegistryValue( "test1", ValueType.REG_SZ,
                REGVALUE_SZ );
        r.setValue( v );
        assertEquals( "REG_SZ strings must match!", REGVALUE_SZ, r.getValue(
                "test1" ).getData() );

        // test REG_EXPAND_SZ
        v = new RegistryValue( "test2", ValueType.REG_EXPAND_SZ,
                REGVALUE_EXPAND_SZ );
        r.setValue( v );

        // test REG_DWORD
        v = new RegistryValue( "test3", ValueType.REG_DWORD, REGVALUE_DWORD );
        r.setValue( v );
        assertEquals( "REG_DWORD values must match!",
                REGVALUE_DWORD.intValue(),
                ((Integer) r.getValue( "test3" ).getData()).intValue() );
        // test REG_EXPAND_SZ
        v = new RegistryValue( "test2", ValueType.REG_EXPAND_SZ,
                REGVALUE_EXPAND_SZ );
        r.setValue( v );

        // test REG_DWORD_LITTLE_ENDIAN
        v = new RegistryValue( "test4", ValueType.REG_DWORD_LITTLE_ENDIAN,
                REGVALUE_DWORD_LITTLE_ENDIAN );
        r.setValue( v );
        assertEquals( "REG_DWORD_LITTLE_ENDIAN values must match!",
                REGVALUE_DWORD_LITTLE_ENDIAN.intValue(), ((Integer) r.getValue(
                        "test4" ).getData()).intValue() );

        // test raw binary values
        v = new RegistryValue( "test5", ValueType.REG_BINARY, REGVALUE_BINARY );
        r.setValue( v );
        assertTrue( "REG_BINARY values must match!", Arrays.equals(
                (byte[]) r.getValue( "test5" ).getData(), REGVALUE_BINARY ) );

        // test REG_QWORD
        v = new RegistryValue( "test6", ValueType.REG_QWORD, REGVALUE_QWORD );
        r.setValue( v );

        // test REG_MULTI_SZ
        String[] REGVALUE_MULTI = { "TESTING", "testing", "1", "TWO", "three",
                "4" };

        v = new RegistryValue( "multi", ValueType.REG_MULTI_SZ, REGVALUE_MULTI );
        r.setValue( v );

        String[] multi = (String[]) r.getValue( "multi" ).getData();
        assertEquals( "REG_MULTI_SZ lengths must match!",
                REGVALUE_MULTI.length, multi.length );

        for ( int i = 0; i < multi.length; i++ ) {
            assertEquals( "REG_MULTI_SZ values (" + i + ") must match!",
                    REGVALUE_MULTI[i], multi[i] );

        }

        //test REG_VALUE_QWORD
        v = new RegistryValue( "test6", ValueType.REG_QWORD, REGVALUE_QWORD );
        r.setValue( v );
        assertEquals( "REG_QWORD values must match!",
                REGVALUE_QWORD.longValue(),
                ((Long) r.getValue( "test6" ).getData()).longValue() );

        // test root (null and empty) values
        v = new RegistryValue( "", ValueType.REG_SZ, REGVALUE_SZ );
        r.setValue( v );
        assertTrue( "null and \"\" values do not match!",
                r.getValue( null ).getData().equals( REGVALUE_SZ ) );

    }

    /**
     * Test that the <code>subkeys</code> method correctly returns an Iterator.
     */
    public void testSubkeys() {
        RegistryKey.initialize();

        RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
        assertNotNull( "subkeys() must not be null!", r.subkeys() );
    }

    /**
     * Tests that the <code>subkeys</code> method correctly iterates through subkeys.
     */
    public void testKeyIterator() {
        RegistryKey.initialize();

        int NUM_KEYS = 10;

        RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
        if ( r.exists() ) {
            r.delete();
        }

        r.create();
        for ( int index = 0; index < NUM_KEYS; index++ ) {
            r.createSubkey( "subkey-" + index );
        }

        if ( r.hasSubkeys() ) {
            System.out.println( "In key " + r.toString() );
            int count = 0;
            Iterator i = r.subkeys();
            while ( i.hasNext() ) {
                RegistryKey rr = (RegistryKey) i.next();
                System.out.println( "Found subkey: " + rr.toString() );
                count++;
            }

            assertEquals( "We created these keys - they should exist!",
                    NUM_KEYS, count );
        } else {
            fail( "Subkeys should have been created - this is serious!" );
        }
    }

    /**
     * Tests that the <code>values</code> method returns an Iterator.
     */
    public void testValues() {
        RegistryKey.initialize();

        RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
        assertNotNull( "values() must not be null!", r.values() );
    }

    /**
     * Tests that the <code>values</code> method correctly iterates through values.
     */
    public void testValueIterator() {
        RegistryKey.initialize( "jRegistryKey-dev" );

        String REGVALUE_EXPAND_SZ = "PATH=%path%";
        Integer REGVALUE_DWORD = new Integer( 1 );
        Integer REGVALUE_DWORD_LITTLE_ENDIAN = new Integer( 2 );
        byte[] REGVALUE_BINARY = new byte[] { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };

        RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
        if ( r.exists() ) {
            r.delete();
        }
        r.create();

        RegistryValue v = new RegistryValue( "test1", ValueType.REG_SZ,
                REGVALUE_SZ );
        r.setValue( v );

        v = new RegistryValue( "test2", ValueType.REG_EXPAND_SZ,
                REGVALUE_EXPAND_SZ );
        r.setValue( v );

        v = new RegistryValue( "test3", ValueType.REG_DWORD, REGVALUE_DWORD );
        r.setValue( v );

        v = new RegistryValue( "test4", ValueType.REG_DWORD_LITTLE_ENDIAN,
                REGVALUE_DWORD_LITTLE_ENDIAN );
        r.setValue( v );

        v = new RegistryValue( "test5", ValueType.REG_BINARY, REGVALUE_BINARY );
        r.setValue( v );

        // test 6
        v = new RegistryValue( null, ValueType.REG_SZ, REGVALUE_SZ );
        r.setValue( v );

        if ( r.hasValues() ) {
            System.out.println( "In key " + r.toString() );
            int count = 0;
            Iterator i = r.values();
            while ( i.hasNext() ) {
                RegistryValue rv = (RegistryValue) i.next();
                System.out.println( "Found value: " + rv.toString() );
                count++;
            }

            assertEquals( "We created these keys - they should exist!", 6,
                    count );
        } else {
            fail( "Values should have been created - this is serious!" );
        }
    }
    
    /**
     * Tests REG_EXPAND_SZ reads are not expanded
     */
    public void testExpandBypass() {
        RegistryKey.initialize();

        RegistryKey r = new RegistryKey( RootKey.HKEY_CURRENT_USER, TESTKEY );
        RegistryValue v = new RegistryValue( "EXPAND", ValueType.REG_EXPAND_SZ,
                "%systemroot%" );
        r.setValue( v );
        
        assertEquals( "REG_EXPAND_SZ values should match, but may vary by system!",
                "C:\\WINDOWS",
                r.getValue( "EXPAND" ).getData() );

        r.bypassExpansion(true);
        assertEquals( "REG_EXPAND_SZ values must match!",
        		"%systemroot%",
                r.getValue( "EXPAND" ).getData() );

        r.bypassExpansion(false);
        assertEquals( "REG_EXPAND_SZ values must match original values!",
                "C:\\WINDOWS",
                r.getValue( "EXPAND" ).getData() );

    }
}