/* RegistryKey.cpp
 *
 * jRegistryKey - A JNI wrapper of the Windows Registry functions.
 * Copyright (c) 2001 - 2005 BEQ Technologies Inc.
 * 9407 - 27 Avenue
 * Edmonton, Alberta
 * T6N - 1L9 Canada
 * (780) 430-0056
 * (780) 437-6121 (fax)
 * http://www.beq.ca
 *
 * Original Author: Joe Robinson <joe@beq.ca>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation; either version 2.1 of the License, or (at your
 * option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author BEQ Technologies Inc.
 * @version $Revision: 1.28 $
 * 
 * $Id: RegistryKey.cpp,v 1.28 2007/04/28 05:46:46 bayblade Exp $
 * 
 * $Log: RegistryKey.cpp,v $
 * Revision 1.28  2007/04/28 05:46:46  bayblade
 * Fix for Bug 1699511 as proposed by Tammo Riedinger
 *
 * Revision 1.27  2007/03/04 02:28:11  bayblade
 * New indentation.  Fix for bug 1663669
 *
 * Revision 1.26  2006/11/15 02:37:55  bayblade
 * Injected some dynamic malloc values so MSVC won't puke on the source
 *
 * Revision 1.25  2006/07/10 01:41:40  bayblade
 * Remote Registry and last access time goodness, courtesy Dave Kipfer <kipfer@cae.com>
 *
 * Revision 1.24  2006/07/06 05:15:51  bayblade
 * Simple cleanup of extractPath() and handful of unused jstring path references
 *
 * Revision 1.23  2006/06/11 16:46:53  bayblade
 * Introducing Win32Exception ...
 * Subclassed RegistryException with new Win32Exception that contains the windows error message, and a method to see the error code as an int rather than a formatted string containing the error code.  The exception message is different, which may break logic that parses this message, but explicitly catching the subclass now allows for greater refinement when processing what actually broke, and is a better solution overall.
 *
 * Revision 1.22  2006/03/02 03:27:57  bayblade
 * Made sure null is actually NULL
 *
 * Revision 1.21  2006/03/02 03:24:25  bayblade
 * Sanity checks for null value types in both the Java and native code, courtesy Arthur Shagall
 *
 * Revision 1.20  2005/12/05 03:18:59  bayblade
 * Hacked to accomodate the versioning sync between the jni and the java portions of the code
 *
 * Revision 1.19  2005/05/10 01:01:46  bayblade
 * 3 things happened here: added support for reg_qword, support for reg_multi_sz and finally unhooped unicode character conversion.
 *
 * Revision 1.18  2005/03/01 00:16:11  bayblade
 * Filtered via indent, removed compiler warnings and other trivial "cleansing"
 *
 * Revision 1.17  2005/02/25 22:09:11  bayblade
 * Fix to bug report 1151552.  We now allow "" (and null) for keynames (default name for keys).  While not rigorously tested for jni problems, it does seem to work for simple cases.
 *
 * Revision 1.16  2005/02/03 00:08:41  bayblade
 * Cleaned up remnants from recent bug fixes.
 *
 * Revision 1.15  2005/01/27 19:31:55  bayblade
 * Fix to support request 1110516
 *
 * Revision 1.14  2005/01/25 22:41:28  bayblade
 * Most noticaly, the fix for 1105000 which both reworks the Unicode to UTF-8 mechanism and permits REG_SZ data to run through the system as byte[] (though byte[] defaults to REG_BINARY)
 *
 * Revision 1.13  2004/10/28 19:30:21  bayblade
 * Added and began using jstringToWindows() method by David Wendt (wendt@us.ibm.com)
 *
 * Revision 1.12  2004/08/10 17:55:22  bayblade
 * Fixed bug [ 1004614 ] Error deleting value using registryKey.deleteValue()
 *
 * Revision 1.11  2004/04/12 21:37:44  mollinie
 * Added a native linking test method.
 *
 * Revision 1.10  2004/04/07 18:57:23  mollinie
 * Renamed member fields according to our conventions.
 *
 * Revision 1.9  2004/04/07 18:24:56  mollinie
 * Added poor man's support for extra registry data types, including REG_MULTI_SZ, REG_LINK, REG_FULL_RESOURCE_DESCRIPTOR and REG_RESOURCE_REQUIREMENTS_LIST.
 *
 * Revision 1.8  2004/04/06 21:27:17  mollinie
 * Added support for international strings.
 *
 * Revision 1.7  2004/04/06 18:19:49  mollinie
 * Reformatted function setValue.
 *
 * Revision 1.6  2004/04/05 21:09:40  mollinie
 * Reformated getValue and extracted functions extractDataValue (create dataValue from native data) and nameType (associates a string to a registry data type).
 *
 * Revision 1.5  2004/04/03 00:15:39  mollinie
 * Extracted the path extraction code (common in openKey, delete and create) in function extractPath.
 *
 * Revision 1.4  2004/04/02 23:54:47  mollinie
 * Extracted the common code to extract rootKey and key path and open key into a common method openKey.
 * Moved utility functions prototypes to header jRegistryKey.h
 *
 * Revision 1.3  2004/04/02 20:52:36  mollinie
 * Changed the return type of getRootKey from int to HKEY and reflected the change in the code.
 * Added missing sanity checks in KeyIterator.initializeFields and ValueIterator.initializeFields.
 *
 * Revision 1.2  2004/04/02 20:35:34  mollinie
 * Reorganized the code of all "small" functions according to structured programing principles.
 * Added missing sanity checks and removed useless ones.
 *
 * Revision 1.1  2004/03/31 23:16:22  mollinie
 * Re-organized the native code organization and build process.
 *
 * Revision 1.2  2004/03/29 19:48:54  mollinie
 * Preliminary reformating.
 *
 * Revision 1.1.1.1  2004/03/25 22:05:38  bayblade
 * Initial Import
 *
 *
 */

// relinquish all the stupidity that is ansi and utf-8
#define UNICODE 1

#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include <shlwapi.h>

#include "jniClassDescriptor.h"
#include "jniFieldDescriptor.h"
#include "version.h"
#include "jRegistryKey.h"
#include "ca_beq_util_win32_registry_RegistryKey.h"

/**
 * ::JNI_OnLoad()
 *    Invoked by the Java Virtual Machine when the native library is loaded by
 *    the System.loadLibrary() call.
 * @return
 */
JNIEXPORT jint JNICALL
JNI_OnLoad ( JavaVM * jvm, void *reserved ) {
	JNIEnv *env;
	jint jniVersion = JNI_ERR;

	if ( jvm->GetEnv ( ( void ** ) &env, JNI_VERSION_1_2 ) == JNI_OK ) {
		jclass cls = env->FindClass ( registryDescriptor::jcd_RegistryKey );
		if ( cls != NULL ) {
			jniVersion = JNI_VERSION_1_2;
		}
	}
	return jniVersion;
}

/**
 * JNI_OnUnload()
 *    Invoked by the Java Virtual Machine when the native library is unloaded.
 */
JNIEXPORT void JNICALL
JNI_OnUnload ( JavaVM * jvm, void *reserved ) {
	return;
}

/**
 * ::getRootKey()
 *   Helper function that retreives the root key from a RegistryKey object.
 */
HKEY getRootKey(JNIEnv * env, jobject obj) {
	HKEY rootKey = 0;

	jclass cls_RegistryKey = env->GetObjectClass(obj);
	jfieldID fid_root = env->GetFieldID(cls_RegistryKey, "m_root",
			registryDescriptor::jfd_RootKey);
	if (fid_root != NULL) {
		jobject obj_RootKey = env->GetObjectField(obj, fid_root);
		jclass cls_RootKey = env->GetObjectClass(obj_RootKey);

		jmethodID mid_getValue = env->GetMethodID(cls_RootKey, "getValue",
				"()I");
		if (mid_getValue != NULL) {
			rootKey = ( HKEY ) env->CallIntMethod(obj_RootKey, mid_getValue);
		}
	}
	return rootKey;
}

/**
 * ::extractMachine()
 *   Helper function that extracts the machine portion of a registry key path as a char*
 */
const wchar_t * extractMachine(JNIEnv * env, jobject obj,
		jstring * jReferenceMachine) {
	const wchar_t *wMachine = NULL;

	jclass cls_RegistryKey = env->GetObjectClass(obj);
	jfieldID fid_machine = env->GetFieldID(cls_RegistryKey, "m_Machine",
			fieldDescriptor::jfd_String);
	if (fid_machine != NULL) {
		jstring jMachine = ( jstring ) env->GetObjectField(obj, fid_machine);
		if (jReferenceMachine != NULL) {
			*jReferenceMachine = jMachine;
		}
		wMachine = jstringToWindows(env, jMachine);
	}
	return wMachine;
}
/**
 * ::extractPath()
 *   Helper function that extracts a registry key path as a char*
 */
const wchar_t * extractPath(JNIEnv * env, jobject obj, jstring * jReferencePath) {
	const wchar_t * wPath = NULL;

	jclass cls_RegistryKey = env->GetObjectClass(obj);
	jfieldID fid_path = env->GetFieldID(cls_RegistryKey, "m_path",
			fieldDescriptor::jfd_String);
	if (fid_path != NULL) {
		jstring path = ( jstring ) env->GetObjectField(obj, fid_path);
		if (jReferencePath != NULL) {
			*jReferencePath = path;
		}
		wPath = jstringToWindows(env, path);

	}
	return wPath;
}

/**
 * ::jstringToWindows
 * 
 * Contrary to what the examples of American English using idiot programmer 
 * demonstrates, nether the Sun nor the Microsoft UTF character-conversion 
 * routines do what they are supposed to any universal concencus of what the 
 * 8-bit should be.
 * The fundamental flaw these programmers overlook is that both MS and Sun are 
 * inherently 16-bit under the hood and then go on to give you nonsensical 
 * conversion routines for 8-bit characters.  
 * Now if we stick to 16 bits, we discover unicode is unicode is unicode.
 */
const wchar_t * jstringToWindows(JNIEnv * env, jstring jstr) {
	if (jstr == NULL) {
		return NULL;
	}

	jboolean copy = JNI_TRUE;
	return ( wchar_t * ) env->GetStringChars(jstr, &copy);

}

/**
 * ::windowsToJString
 *   Home-brewed equivelency method to accomplish the opposite of jstringToWindows
 */
const jstring windowsToJString(JNIEnv * env, wchar_t * str) {
	return env->NewString( ( jchar * ) str, wcslen(str) );
}

/**
 * ::freeWindowsFromJString
 *   return memory allocated by jstringToWindows to its appropriate heap
 */
void freeWindowsFromJString(JNIEnv * env, jstring jstr, const wchar_t * str) {
	env->ReleaseStringChars(jstr, (jchar*) str);
}

void throwWin32(JNIEnv * env, jint status) {
	jclass cls_Win32Exception =
			env->FindClass(registryDescriptor::jcd_Win32Exception);

	jmethodID mid = env->GetMethodID(cls_Win32Exception, "<init>", "(I)V");

	if (mid == NULL) {
		// we're really hooped here, consider something more drastic
		return;
	}

	jthrowable exception = (jthrowable) env->NewObject(cls_Win32Exception, mid,
			status);

	env->Throw(exception);
}

/**
 * ::openKey()
 * Helper function that opens the key represented by the RegistryKey object
 * @param obj MUST be a RegistryKey object.
 * @return a handle to the opened key.
 */
HKEY openKey(JNIEnv *env, jobject obj, REGSAM samDesired) {
	HKEY hkey = NULL;

	HKEY localRootKey = getRootKey(env, obj);
	HKEY rootKey = localRootKey;
	jstring jMachine;
	jstring jPath;
	const wchar_t *wMachine = extractMachine(env, obj, &jMachine);
	const wchar_t *wPath = extractPath(env, obj, &jPath);

	if ( (rootKey != NULL )) {

		if (wMachine != NULL) {
			LONG status = RegConnectRegistry(wMachine, localRootKey, &rootKey);
			if (status != ERROR_SUCCESS) {
				throwWin32(env, status);
				rootKey = localRootKey;
			}
		}

		LONG status = RegOpenKeyEx(rootKey, wPath, 0, samDesired, &hkey);
		if (status != ERROR_SUCCESS) {
			throwWin32(env, status);
			hkey = NULL;
		}

		// Room for all kinds of headaches, given typical MS crap.  Watch closely
		if (rootKey != localRootKey) {
			RegCloseKey(rootKey);
		}
	}

	if (wPath != NULL) {
		freeWindowsFromJString(env, jPath, wPath);
	}
	if (wMachine != NULL) {
		freeWindowsFromJString(env, jMachine, wMachine);
	}

	return hkey;
}

/**
 * RegistryKey::testInitialized()
 *   Does nothing. If the native library was not initialized correctly,
 *   a call to this method (as well as any other) will throw an
 *   UnsatisfiedLinkError.
 */
JNIEXPORT void JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_testInitialized ( JNIEnv * env,
		jclass cls ) {}

JNIEXPORT jstring JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_getJNIVersion ( JNIEnv * env, jclass cls ) {
	return windowsToJString( env, (wchar_t *) CURRENT_VERSION );
}

/**
 * RegistryKey::exists()
 *    Tests if this registry key, as defined by the RootKey and path, currently
 *    exists in the system registry.
 */
JNIEXPORT jboolean JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_exists ( JNIEnv * env, jobject obj ) {
	bool keyExists = false;
	HKEY hkey = openKey ( env, obj, KEY_READ );
	if ( hkey != NULL ) {
		RegCloseKey( hkey );
		keyExists = true;
	} else {
		// an exception has been thrown when opening the key failed, but we do not
		// want this behavior in the method. Clear the exception.
		env->ExceptionClear();
	}
	return keyExists;
}

/**
 * RegistryKey::create()
 *    Creates this registry key in the system registry.
 */
JNIEXPORT void JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_create ( JNIEnv * env, jobject obj ) {
	HKEY rootKey = getRootKey ( env, obj );
	jstring jPath;
	const wchar_t *szPath = extractPath ( env, obj, &jPath );
	if ( rootKey != 0 ) {
		HKEY hkey;
		DWORD dwDisposition;
		LONG status = RegCreateKeyEx ( rootKey, szPath, 0, NULL,
				REG_OPTION_NON_VOLATILE,
				KEY_ALL_ACCESS, NULL, &hkey,
				&dwDisposition );
		if ( status == ERROR_SUCCESS ) {
			RegCloseKey ( hkey );
			if ( dwDisposition == REG_OPENED_EXISTING_KEY ) {
				const int msg_len = 200;
				char msg[msg_len + 1];
				_snprintf ( msg, msg_len,
						"Attempted to create existing key %s.", szPath );
				jclass cls_RegistryException =
				env->
				FindClass ( registryDescriptor::jcd_RegistryException );
				env->ThrowNew ( cls_RegistryException, msg );
			}
		} else {
			throwWin32( env, status );
		}
	}

	if ( szPath != NULL ) {
		freeWindowsFromJString ( env, jPath, szPath );
	}
}

/**
 * RegistryKey::delete()
 *    Deletes this registry key from the system registry.  The key must not
 *    possess either values or subkeys.
 */
JNIEXPORT void JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_delete ( JNIEnv * env, jobject obj ) {
	HKEY rootKey = getRootKey ( env, obj );
	jstring jPath;
	const wchar_t *wPath = extractPath ( env, obj, &jPath );
	if ( rootKey != 0 ) {
		LONG status = SHDeleteKey ( rootKey, wPath );
		if ( status != ERROR_SUCCESS ) {
			throwWin32( env, status );
		}
	}
	if ( wPath != NULL ) {
		freeWindowsFromJString ( env, jPath, wPath );
	}

}

/**
 * RegistryKey::getLastModifiedFileTime()
 *    Returns the last modified time of this key as a Windows FILETIME
 */
JNIEXPORT jlong JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_getLastModifiedFileTime ( JNIEnv * env,
		jobject obj ) {
	jlong wLastModifiedTime = 0;
	HKEY hkey = openKey ( env, obj, KEY_READ );
	if ( hkey != NULL ) {
		FILETIME wTime;
		LONG status = RegQueryInfoKey ( hkey, NULL, NULL, NULL, NULL, NULL, NULL,
				NULL, NULL, NULL, NULL, &wTime );
		RegCloseKey ( hkey );
		if ( status == ERROR_SUCCESS ) {
			wLastModifiedTime = wTime.dwHighDateTime;
			wLastModifiedTime <<= 32;
			wLastModifiedTime |= wTime.dwLowDateTime;
		} else {
			throwWin32( env, status );
		}
	}
	return wLastModifiedTime;
}

/**
 * RegistryKey::hasSubkeys()
 *    Tests if this registry key possesses subkeys.  Use subkeys() to retrieve
 *    an iterator for available subkeys.
 */
JNIEXPORT jboolean JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_hasSubkeys ( JNIEnv * env,
		jobject obj ) {
	bool hasSubkeys = false;

	HKEY hkey = openKey ( env, obj, KEY_READ );
	if ( hkey != NULL ) {
		DWORD count;
		LONG status =
		RegQueryInfoKey ( hkey, NULL, NULL, NULL, &count, NULL, NULL,
				NULL,
				NULL, NULL, NULL, NULL );
		if ( status == ERROR_SUCCESS ) {
			RegCloseKey ( hkey );
			if ( count> 0 ) {
				hasSubkeys = true;
			}
		} else {
			throwWin32( env, status );
		}
	}
	return hasSubkeys;
}

/**
 * RegistryKey::hasValues()
 *    Tests if this registry key possess any values.
 */
JNIEXPORT jboolean JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_hasValues ( JNIEnv * env,
		jobject obj ) {
	bool hasValues = false;

	HKEY hkey = openKey ( env, obj, KEY_READ );
	if ( hkey != NULL ) {
		DWORD count;
		LONG status = RegQueryInfoKey ( hkey, NULL, NULL, NULL, NULL, NULL, NULL,
				&count, NULL, NULL, NULL, NULL );
		if ( status == ERROR_SUCCESS ) {
			RegCloseKey ( hkey );
			if ( count> 0 ) {
				hasValues = true;
			}
		} else {
			throwWin32( env, status );
		}
	}
	return hasValues;
}

/**
 * RegistryKey::hasValue()
 *    Tests if this registry key possess the specified value.
 */
JNIEXPORT jboolean JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_hasValue ( JNIEnv * env,
		jobject obj,
		jstring name ) {
	bool hasValue = false;

	HKEY hkey = openKey ( env, obj, KEY_READ );
	if ( hkey != NULL ) {
		const wchar_t * wName = jstringToWindows ( env, name );
		if ( RegQueryValueEx ( hkey, wName, NULL, NULL, NULL, NULL ) ==
				ERROR_SUCCESS ) {
			hasValue = true;
		}

		if ( wName != NULL ) {
			freeWindowsFromJString ( env, name, wName );
		}
		RegCloseKey ( hkey );
	}
	return hasValue;
}

/**
 * ::nameType()
 *   Helper function that associates a registry type with a string
 *   representation thereof.
 * @return
 */
char * nameType(JNIEnv * env, DWORD type) {
	switch (type) {
	case REG_NONE:
		return "REG_NONE";
	case REG_BINARY:
		return "REG_BINARY";
	case REG_DWORD_LITTLE_ENDIAN:
		return "REG_DWORD_LITTLE_ENDIAN";
	case REG_DWORD_BIG_ENDIAN:
		return "REG_DWORD_BIG_ENDIAN";
	case REG_SZ:
		return "REG_SZ";
	case REG_EXPAND_SZ:
		return "REG_EXPAND_SZ";
	case REG_MULTI_SZ:
		return "REG_MULTI_SZ";
	case REG_LINK:
		return "REG_LINK";
	case REG_RESOURCE_LIST:
		return "REG_RESOURCE_LIST";
	case REG_FULL_RESOURCE_DESCRIPTOR:
		return "REG_FULL_RESOURCE_DESCRIPTOR";
	case REG_RESOURCE_REQUIREMENTS_LIST:
		return "REG_RESOURCE_REQUIREMENTS_LIST";
	case REG_QWORD:
		return "REG_QWORD";
	default:
		const size_t msg_len = 200;
		char msg[msg_len + 1];
		_snprintf(msg, msg_len, "Unrecognized registry field type number: %ld",
				type);
		jclass cls_RegistryException =
				env->FindClass(registryDescriptor::jcd_RegistryException);
		env->ThrowNew(cls_RegistryException, msg);
		return NULL;
	}
}

/**
 * ::makeDataValue()
 * Helper function that creates a Java object from its registry equivalent
 * @return null for unhandled values, some other value, as appropriate
 */
jobject makeDataValue(JNIEnv * env, DWORD type, DWORD size, BYTE * data) {
	jobject dataValue = NULL;
	switch (type) {
	// raw, unformatted, binary data
	case REG_NONE:
	case REG_BINARY:
		// as well as un-handled data types
	case REG_LINK:
	case REG_RESOURCE_LIST:
	case REG_FULL_RESOURCE_DESCRIPTOR:
	case REG_RESOURCE_REQUIREMENTS_LIST:
		dataValue = env->NewByteArray(size);
		if (dataValue != NULL) {
			env->SetByteArrayRegion( ( jbyteArray ) dataValue, 0, size,
					( jbyte * ) data );
			if (env->ExceptionCheck() == JNI_TRUE) {
				dataValue = NULL;
			}
		}
		break;

	case REG_DWORD_BIG_ENDIAN:
	case REG_DWORD_LITTLE_ENDIAN: {
		jclass cls_Integer = env->FindClass(classDescriptor::jcd_Integer);
		if (cls_Integer != NULL) {
			jmethodID mid_IntegerConstructor = env->GetMethodID(cls_Integer,
					"<init>", "(I)V");
			if (mid_IntegerConstructor != NULL) {
				dataValue = env->NewObject(cls_Integer, mid_IntegerConstructor,
						*( jint * ) data );
			}
		}
		break;
	}
	case REG_QWORD: {
		jclass cls_Long = env->FindClass(classDescriptor::jcd_Long);
		if (cls_Long != NULL) {
			jmethodID mid_LongConstructor = env->GetMethodID(cls_Long,
					"<init>", "(J)V");
			if (mid_LongConstructor != NULL) {
				dataValue = env->NewObject(cls_Long, mid_LongConstructor,
						*( jlong * ) data );
			}
		}
		break;
	}
	case REG_EXPAND_SZ: {

		if (size == 0) {
			dataValue = windowsToJString(env, ( wchar_t * )L"" );
			break;
		}
		DWORD required = ExpandEnvironmentStrings ( ( wchar_t * ) data, NULL, 0 );
		if ( required != 0 ) {
			// accomodate the null at the end
			wchar_t* tmp = (wchar_t*)malloc((required + 1) * sizeof(wchar_t));
			// new wchar_t[required + 1];
			DWORD copied =
			ExpandEnvironmentStrings ( ( wchar_t * ) data, tmp, required );
			if ( copied != 0 ) {
				dataValue = windowsToJString ( env, tmp );
			} else {
				const size_t msg_len = 200;
				char msg[msg_len + 1];
				_snprintf ( msg, msg_len, "Error expanding registry string \"%s\"", data );
				jclass cls_RegistryException = env-> FindClass ( registryDescriptor::jcd_RegistryException );
				env->ThrowNew ( cls_RegistryException, msg );
			}
			free(tmp);
			//delete tmp;
		}
		break;

	}
	case REG_SZ: {
		if (size == 0) {
			dataValue = windowsToJString(env, ( wchar_t * )L"" );
			break;
		}
		dataValue = windowsToJString ( env, ( wchar_t * ) data );
		break;
	}
	case REG_MULTI_SZ: {
		unsigned int i;
		int retIndex = 0;
		int count = 0;
		int start = 0;

		for ( i = 0; i < size; i += sizeof ( wchar_t ) ) {
			if ( ( wchar_t ) data[i] == 0 ) {
				count++;
			}
		}
		dataValue = env->NewObjectArray ( count, env->FindClass ( classDescriptor::jcd_String ), NULL );
		for ( i = 0; i < size; i += sizeof ( wchar_t ) ) {
			if ( ( wchar_t ) data[i] == 0 ) {
				jstring str = windowsToJString ( env,( wchar_t * ) &data[start] );
				env->SetObjectArrayElement ( ( jobjectArray ) dataValue, retIndex++, str );
				start = i + sizeof ( wchar_t );
			}
		}
		break;
	}
	default: {
		const size_t msg_len = 200;
		char msg[msg_len + 1];
		_snprintf ( msg, msg_len, "Unsupported data type number: %ld", type );
		jclass cls_RegistryException =
		env->FindClass ( registryDescriptor::jcd_RegistryException );
		env->ThrowNew ( cls_RegistryException, msg );
		break;
	}
}

return dataValue;
}

/**
 * RegistryKey::getValue()
 *    Returns a RegistryValue representing the specified value.
 */
JNIEXPORT jobject JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_getValue ( JNIEnv * env,
		jobject obj,
		jstring name ) {
	jobject value = NULL;

	HKEY hkey = openKey ( env, obj, KEY_READ );

	const wchar_t *szName = jstringToWindows ( env, name );

	if ( ( hkey != NULL ) ) {
		DWORD type;
		DWORD size;
		LONG status =
		RegQueryValueEx ( hkey, szName, NULL, &type, NULL, &size );
		if ( status == ERROR_SUCCESS ) {
			char *typeName = nameType ( env, type ); // set to string version of registry type
			if ( typeName != NULL ) {
				BYTE *data = ( BYTE * ) malloc ( size );
				if ( data != NULL ) {
					status =
					RegQueryValueEx ( hkey, szName, NULL, &type, data,
							&size );
					if ( status == ERROR_SUCCESS ) {
						jclass cls_RegistryValue = env-> FindClass ( registryDescriptor::jcd_RegistryValue );

						if ( cls_RegistryValue != NULL ) {
							jmethodID mid_Constructor = env->GetMethodID ( cls_RegistryValue, "<init>", "()V" );
							if ( mid_Constructor != NULL ) {

								value = env->NewObject ( cls_RegistryValue, mid_Constructor );

								// Check that we should not be ignoring REG_EXPAND_SZ
								jboolean ignore_expand = JNI_FALSE;
								jclass cls_RegistryKey = env-> FindClass( registryDescriptor::jcd_RegistryKey );

								if (cls_RegistryKey != NULL) {
									jfieldID fid_type = env->GetFieldID( cls_RegistryKey,
											"m_ignore_expand", fieldDescriptor::jfd_boolean );
									if (fid_type != NULL) {
										ignore_expand = env->GetBooleanField( obj, fid_type );
									}
								}
								jobject dataValue;
								if (JNI_TRUE == ignore_expand && REG_EXPAND_SZ == type ) {
									dataValue = makeDataValue( env, REG_SZ, size, data );
								} else {
									dataValue = makeDataValue( env, type, size, data );
								}

								jobject typeValue = NULL;
								jclass cls_ValueType = env->FindClass( registryDescriptor::jcd_ValueType );

								if ( cls_ValueType != NULL ) {
									jfieldID fid_type = env->GetStaticFieldID( 
											cls_ValueType, typeName, registryDescriptor::jfd_ValueType );
									if ( fid_type != NULL ) {
										typeValue = env->GetStaticObjectField( cls_ValueType, fid_type );
									}
								}

								jmethodID mid_setName =
								env->GetMethodID ( cls_RegistryValue,
										"setName",
										"(Ljava/lang/String;)V" );
								jmethodID mid_setType =
								env->GetMethodID ( cls_RegistryValue,
										"setType",
										"(Lca/beq/util/win32/registry/ValueType;)V" );
								jmethodID mid_setData =
								env->GetMethodID ( cls_RegistryValue,
										"setData",
										"(Ljava/lang/Object;)V" );

								if ( ( dataValue != NULL )
										&& ( typeValue != NULL )
										&& ( mid_setName != NULL )
										&& ( mid_setType != NULL )
										&& ( mid_setData != NULL ) ) {
									env->CallVoidMethod ( value, mid_setName, name );
									env->CallVoidMethod ( value, mid_setType, typeValue );
									env->CallVoidMethod ( value, mid_setData, dataValue );
								}
							}
						}
					} else {
						throwWin32( env, status );
					}
					free ( data );
				} else {
					jclass cls_OutOfMemoryError = env-> FindClass ( classDescriptor::jcd_OutOfMemoryError );
					env->ThrowNew ( cls_OutOfMemoryError, "Could not allocate memory." );
				}
			}
		} else {
			throwWin32( env, status );
		}
		RegCloseKey ( hkey );
	}
	if ( szName != NULL ) {
		freeWindowsFromJString ( env, name, szName );
	}

	return value;
}

/**
 * RegistryKey::setValue()
 *    Sets the properties of a registry value according to the properties of
 *    the specified RegistryValue.  If the specified value exists, it will be
 *    modified; if not, it will be created.
 */
JNIEXPORT void JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_setValue ( JNIEnv * env,
		jobject obj,
		jobject value ) {
	HKEY hkey = openKey ( env, obj, KEY_ALL_ACCESS );
	if ( hkey != NULL ) {
		jclass cls_RegistryValue = env->GetObjectClass ( value );
		if ( cls_RegistryValue != NULL ) {
			jstring name = NULL;
			jmethodID mid_getName =
			env->GetMethodID ( cls_RegistryValue, "getName",
					"()Ljava/lang/String;" );
			if ( mid_getName != NULL ) {
				name = ( jstring ) env->CallObjectMethod ( value, mid_getName );
			}

			jobject type = NULL;
			jmethodID mid_getType =
			env->GetMethodID ( cls_RegistryValue, "getType",
					"()Lca/beq/util/win32/registry/ValueType;" );
			if ( mid_getType != NULL ) {
				type = env->CallObjectMethod ( value, mid_getType );
			}

			jobject data = NULL;
			jmethodID mid_getData =
			env->GetMethodID ( cls_RegistryValue, "getData",
					"()Ljava/lang/Object;" );
			if ( mid_getData != NULL ) {
				data = env->CallObjectMethod ( value, mid_getData );
			}

			DWORD typeValue = ( DWORD ) - 1;
			jclass cls_ValueType = NULL;
			if (type != NULL) {
				cls_ValueType = env->GetObjectClass ( type );
			}
			if ( cls_ValueType != NULL ) {
				jmethodID mid_getValue =
				env->GetMethodID ( cls_ValueType, "getValue", "()I" );
				if ( mid_getValue != NULL ) {
					typeValue = env->CallIntMethod ( type, mid_getValue );
				}
			}

			if ( ( type != NULL ) && ( data != NULL )
					&& ( typeValue != ( DWORD ) - 1 ) ) {
				const wchar_t *szName = jstringToWindows ( env, name );

				switch ( typeValue ) {
					case REG_NONE:
					case REG_BINARY: {
						jclass cls_ByteArray =
						env->FindClass ( fieldDescriptor::jfd_byteArray );
						jclass dataClass = env->GetObjectClass ( data );
						if ( ( cls_ByteArray != NULL )
								&& ( dataClass != NULL )
								&& ( env->
										IsAssignableFrom ( dataClass,
												cls_ByteArray ) ==
										JNI_TRUE ) ) {
							BYTE *dataValue =
							( BYTE * ) env->
							GetByteArrayElements ( ( jbyteArray ) data,
									NULL );
							if ( dataValue != NULL ) {
								DWORD size =
								env->GetArrayLength ( ( jarray ) data );
								LONG status = RegSetValueEx ( hkey, szName, 0,
										typeValue,
										dataValue, size );
								if ( status != ERROR_SUCCESS ) {
									const size_t msg_len = 200;
									char msg[msg_len + 1];
									_snprintf ( msg, msg_len,
											"Attempt to set value \"%s\" of type %s failed with error code %ld",
											szName, nameType ( env,
													typeValue ),
											status );
									jclass cls_RegistryException =
									env->
									FindClass ( registryDescriptor::
											jcd_RegistryException );
									env->ThrowNew ( cls_RegistryException,
											msg );
								}
								env->ReleaseByteArrayElements ( ( jbyteArray )
										data,
										( signed char
												* )
										dataValue, 0 );
							}
						} else {
							jclass cls_RegistryException =
							env->
							FindClass ( registryDescriptor::
									jcd_RegistryException );
							env->ThrowNew ( cls_RegistryException,
									"Incompatible data types" );
						}
					}
					break;

					case REG_QWORD: {
						jclass cls_Long =
						env->FindClass ( classDescriptor::jcd_Long );
						jclass dataClass = env->GetObjectClass ( data );
						if ( ( cls_Long != NULL )
								&& ( dataClass != NULL )
								&& ( env->IsAssignableFrom ( dataClass, cls_Long )
										== JNI_TRUE ) ) {
							jmethodID mid_longValue =
							env->GetMethodID ( env->GetObjectClass ( data ),
									"longValue", "()J" );
							if ( mid_longValue != NULL ) {
								jlong lvalue = env->CallLongMethod ( data,
										mid_longValue );
								BYTE *dataValue = ( BYTE * ) & lvalue;
								int size = 8;
								LONG status = RegSetValueEx ( hkey, szName, 0,
										typeValue,
										dataValue, size );
								if ( status != ERROR_SUCCESS ) {
									const size_t msg_len = 200;
									char msg[msg_len + 1];
									_snprintf ( msg, msg_len,
											"Attempt to set value \"%s\" of type %s failed with error code %ld",
											szName, nameType ( env,
													typeValue ),
											status );
									jclass cls_RegistryException =
									env->
									FindClass ( registryDescriptor::
											jcd_RegistryException );
									env->ThrowNew ( cls_RegistryException,
											msg );
								}
							}
						} else {
							jclass cls_RegistryException =
							env->
							FindClass ( registryDescriptor::
									jcd_RegistryException );
							env->ThrowNew ( cls_RegistryException,
									"Incompatible data types" );
						}
					}
					break;

					case REG_DWORD_BIG_ENDIAN:
					case REG_DWORD_LITTLE_ENDIAN: {
						jclass cls_Integer =
						env->FindClass ( classDescriptor::jcd_Integer );
						jclass dataClass = env->GetObjectClass ( data );
						if ( ( cls_Integer != NULL )
								&& ( dataClass != NULL )
								&& ( env->
										IsAssignableFrom ( dataClass, cls_Integer )
										== JNI_TRUE ) ) {
							jmethodID mid_intValue =
							env->GetMethodID ( env->GetObjectClass ( data ),
									"intValue", "()I" );
							if ( mid_intValue != NULL ) {
								int ivalue = env->CallIntMethod ( data,
										mid_intValue );
								BYTE *dataValue = ( BYTE * ) & ivalue;
								int size = 4;
								LONG status = RegSetValueEx ( hkey, szName, 0,
										typeValue,
										dataValue, size );
								if ( status != ERROR_SUCCESS ) {
									throwWin32( env, status );
								}
							}
						} else {
							jclass cls_RegistryException =
							env->
							FindClass ( registryDescriptor::
									jcd_RegistryException );
							env->ThrowNew ( cls_RegistryException,
									"Incompatible data types" );
						}
					}
					break;
					case REG_SZ:
					case REG_EXPAND_SZ: {
						jclass cls_String =
						env->FindClass ( classDescriptor::jcd_String );
						jclass cls_ByteArray =
						env->FindClass ( fieldDescriptor::jfd_byteArray );
						jclass dataClass = env->GetObjectClass ( data );
						if ( ( cls_String != NULL )
								&& ( cls_ByteArray != NULL )
								&& ( dataClass != NULL )
						) {
							if ( env->
									IsAssignableFrom ( dataClass,
											cls_String ) == JNI_TRUE ) {

								const wchar_t *dataValue =
								jstringToWindows ( env,
										( jstring ) data );
								int size =
								sizeof ( wchar_t ) *
								( wcslen ( dataValue ) + 1 );
								LONG status = RegSetValueEx ( hkey, szName, 0,
										typeValue,
										( BYTE * )
										dataValue,
										size );
								freeWindowsFromJString ( env,( jstring ) data, dataValue );

								if ( status != ERROR_SUCCESS ) {
									throwWin32( env, status );
								}
							} else if ( env->
									IsAssignableFrom ( dataClass,
											cls_ByteArray )
									== JNI_TRUE ) {
								BYTE *dataValue = ( BYTE * )
								env->GetByteArrayElements ( ( jbyteArray )
										data,
										NULL );
								if ( dataValue != NULL ) {
									DWORD size =
									env->GetArrayLength ( ( jarray ) data );
									LONG status =
									RegSetValueEx ( hkey, szName, 0,
											typeValue,
											dataValue,
											size );
									if ( status != ERROR_SUCCESS ) {
										const size_t msg_len = 200;
										char msg[msg_len + 1];
										_snprintf ( msg, msg_len,
												"Attempt to set value \"%s\" of type %s failed with error code %ld",
												szName, nameType ( env,
														typeValue ),
												status );
										jclass cls_RegistryException =
										env->
										FindClass ( registryDescriptor::
												jcd_RegistryException );
										env->
										ThrowNew
										( cls_RegistryException, msg );
									}
									env->
									ReleaseByteArrayElements ( ( jbyteArray )
											data,
											( signed
													char * )
											dataValue,
											0 );
								}

							}
						} else {
							jclass cls_RegistryException =
							env->
							FindClass ( registryDescriptor::
									jcd_RegistryException );
							env->ThrowNew ( cls_RegistryException,
									"Incompatible data types (Should be byte[] or String)" );

						}
					}
					break;
					case REG_MULTI_SZ: {
						jclass cls_StringArray =
						env->FindClass ( fieldDescriptor::jfd_StringArray );
						jclass cls_ByteArray =
						env->FindClass ( fieldDescriptor::jfd_byteArray );
						jclass dataClass = env->GetObjectClass ( data );
						if ( cls_StringArray != NULL
								&& cls_ByteArray != NULL && dataClass != NULL ) {
							if ( env->
									IsAssignableFrom ( dataClass,
											cls_StringArray ) ==
									JNI_TRUE ) {

								int length =
								env->GetArrayLength ( ( jobjectArray )
										data );
								int insertSize = 0;
								BYTE *insert = NULL;
								int i;
								for ( i = 0; i < length; i++ ) {
									jstring curr = ( jstring )
									env->
									GetObjectArrayElement ( ( jobjectArray )
											data, i );
									const wchar_t *currValue =
									jstringToWindows ( env, curr );
									int currSize =
									sizeof ( wchar_t ) *
									( wcslen ( currValue ) + 1 );
									insert =
									( BYTE * ) realloc ( insert,
											insertSize +
											currSize );
									if ( insert == NULL ) {
										jclass cls_OutOfMemoryError =
										env->
										FindClass ( classDescriptor::
												jcd_OutOfMemoryError );
										env->
										ThrowNew ( cls_OutOfMemoryError,
												"Could not allocate memory." );
										break;

									}
									memcpy ( &insert[insertSize],
											currValue, currSize );
									freeWindowsFromJString ( env, curr, currValue );
									insertSize += currSize;
								}
								if ( insert == NULL ) {
									break;
								}
								LONG status = RegSetValueEx ( hkey, szName, 0,
										typeValue,
										( BYTE * )
										insert,
										insertSize );
								free ( ( void * ) insert );

								if ( status != ERROR_SUCCESS ) {
									throwWin32( env, status );
								}

							} else if ( env->
									IsAssignableFrom ( dataClass,
											cls_ByteArray )
									== JNI_TRUE ) {
								BYTE *dataValue = ( BYTE * )
								env->GetByteArrayElements ( ( jbyteArray )
										data,
										NULL );
								if ( dataValue != NULL ) {
									DWORD size =
									env->GetArrayLength ( ( jarray ) data );
									LONG status =
									RegSetValueEx ( hkey, szName, 0,
											typeValue,
											dataValue,
											size );
									if ( status != ERROR_SUCCESS ) {
										throwWin32( env, status );
									}
									env->
									ReleaseByteArrayElements ( ( jbyteArray )
											data,
											( signed
													char * )
											dataValue,
											0 );
								}

							}
						} else {
							jclass cls_RegistryException =
							env->
							FindClass ( registryDescriptor::
									jcd_RegistryException );
							env->ThrowNew ( cls_RegistryException,
									"Incompatible data types (Should be byte[] or String[])" );

						}
					}
					break;

					default: {
						jclass cls_RegistryException =
						env->
						FindClass ( registryDescriptor::
								jcd_RegistryException );
						env->ThrowNew ( cls_RegistryException,
								"Unsupported value type" );
					}
					break;
				}

				if ( szName != NULL ) {
					freeWindowsFromJString ( env, name, szName );
				}

			}
		}
		RegCloseKey ( hkey );
	}
}

/**
 * RegistryKey::deleteValue()
 *    Deletes the specified value from this registry key.
 */
JNIEXPORT void JNICALL
Java_ca_beq_util_win32_registry_RegistryKey_deleteValue ( JNIEnv * env,
		jobject obj,
		jstring name ) {
	HKEY hkey = openKey ( env, obj, KEY_WRITE );
	if ( hkey != NULL ) {
		const wchar_t *szName = jstringToWindows ( env, name );

		LONG status = RegDeleteValue ( hkey, szName );

		if ( status != ERROR_SUCCESS ) {
			throwWin32( env, status );
		}
		RegCloseKey ( hkey );

		if ( szName != NULL ) {
			freeWindowsFromJString ( env, name, szName );
		}

	}
}
